from nipype.interfaces.base import (traits, Undefined, isdefined)
from nipype.interfaces.io import (IOBase, add_traits)
from nipype.interfaces.utility import IdentityInterface


class Gate(IOBase):
    """
    The Gate interface class is a copy of the Nipype IdentityInterface.
    Since Nipype elides IdentityInterface nodes from the execution graph,
    IdentityInterface cannot be used to constrain execution order, as in
    the example below. Gate is an IdentityInterface look-alike that
    preserves the node connection in the execution graph.

    Example::

        from qipipe.interfaces import Gate
        gate = Node(Gate(fields'['a', 'b']))
        workflow.connect(upstream1, 'a', gate, 'a')
        workflow.connect(upstream2, 'b', gate, 'b')
        workflow.connect(gate, 'a', downstream, 'a')

    In this example, the ``gate`` node starts after both ``upstream1`` and
    ``upstream2`` finish. Consequently, the ``downstream`` node starts only
    after ``upstream2`` finishes. This execution precedence constraint does
    not hold if gate were an IdentityInterface.
    """
    input_spec = IdentityInterface.input_spec
    output_spec = IdentityInterface.output_spec

    def __init__(self, fields=None, mandatory_inputs=True, **inputs):
        super(Gate, self).__init__(**inputs)
        if fields is None or not fields:
            raise ValueError('Gate fields must be a non-empty list')
        # Each input must be in the fields.
        for in_field in inputs:
            if in_field not in fields:
                raise ValueError('Gate input is not in the fields: %s' % in_field)
        self._fields = fields
        self._mandatory_inputs = mandatory_inputs
        add_traits(self.inputs, fields)
        # Adding any traits wipes out all input values set in superclass initialization,
        # even it the trait is not in the add_traits argument. The work-around is to reset
        # the values after adding the traits.
        self.inputs.set(**inputs)

    def _add_output_traits(self, base):
        undefined_traits = {}
        for key in self._fields:
            base.add_trait(key, traits.Any)
            undefined_traits[key] = Undefined
        base.trait_set(trait_change_notify=False, **undefined_traits)
        return base

    def _list_outputs(self):
        #manual mandatory inputs check
        if self._fields and self._mandatory_inputs:
            for key in self._fields:
                value = getattr(self.inputs, key)
                if not isdefined(value):
                    msg = "%s requires a value for input '%s' because it was listed in 'fields'. \
                    You can turn off mandatory inputs checking by passing mandatory_inputs = False to the constructor." % \
                    (self.__class__.__name__, key)
                    raise ValueError(msg)

        outputs = self._outputs().get()
        for key in self._fields:
            val = getattr(self.inputs, key)
            if isdefined(val):
                outputs[key] = val
        return outputs



"""The top-level Quantitative Imaging Pipeline module."""

__version__ = '4.5.5'
"""
The one-based major.minor.patch version.
The version numbering scheme loosely follows http://semver.org/.
The major version is incremented when a significant feature
set is introduced. The minor version is incremented when there
is a functionality change. The patch version is incremented when
there is a refactoring or bug fix. All major, minor and patch
version numbers begin at 1.
"""




from qiutil.which import which

DISTRIBUTABLE = not not which('qsub')
"""
Flag indicating whether the workflow can be distributed over a cluster.
This flag is True if ``qsub`` is in the execution path, False otherwise.
"""



import re
from nipype.interfaces.base import (traits, BaseInterfaceInputSpec,
                                    BaseInterface)
from qiutil.qiprofile_helper import QIProfile


class UpdateQIProfileInputSpec(BaseInterfaceInputSpec):
    project = traits.Str(mandatory=True, desc='The XNAT project id')

    subject = traits.Str(mandatory=True, desc='The XNAT subject name')

    session = traits.Str(mandatory=True, desc='The XNAT session name')

    session_params = traits.Dict({}, usedefault=True,
                                 desc='The imaging acquistion parameters'
                                      ' name-value dictionary')

    resource = traits.Str(desc='The XNAT modeling resource name')

    modeling_params = traits.Dict({}, usedefault=True,
                                  desc='The PK modeling parameters'
                                       ' name-value dictionary')


class UpdateQIProfile(BaseInterface):
    """
    The ``UpdateQIProfile`` Nipype interface updates the Imaging Profile
    database.
    """

    input_spec = UpdateQIProfileInputSpec

    def _run_interface(self, runtime):
        db = QIProfile()
        sess = db.save_session(self.inputs.project, self.inputs.subject,
                               self.inputs.session, **self.inputs.session_params)
        if self.inputs.resource:
            mdl = Modeling(session=sess, name=self.inputs.resource,
                           **self.inputs.modeling_params)
            mdl.save()

        return runtime



"""
Image processing preparation.

The staging package defines the functions used to prepare the study image files
for import into XNAT, submission to the TCIA QIN collections and pipeline
processing.
"""



import os
import glob
import shutil
from nose.tools import (assert_equal, assert_is_not_none)
from ...helpers.logging import logger
from qipipe.staging.map_ctp import CTPPatientIdMap

COLLECTION = 'Sarcoma'
"""The test collection."""

SUBJECTS = ["Sarcoma%03d" % i for i in range(8, 12)]
"""The test subjects."""

PAT = "ptid/(Sarcoma\d{3})\s*=\s*QIN-\w+-\d{2}-(\d{4})"
"""The CTP map pattern specified by the QIN TCIA curator."""


class TestMapCTP(object):

    """Map CTP unit tests."""

    def test_map_ctp(self):
        logger(__name__).debug("Testing Map CTP on %s..." % SUBJECTS)
        ctp_map = CTPPatientIdMap()
        ctp_map.add_subjects(COLLECTION, *SUBJECTS)
        for sbj in SUBJECTS:
            ctp_id = ctp_map.get(sbj)
            assert_is_not_none(ctp_id, "Subject was not mapped: %s" % sbj)
            qin_nbr = int(sbj[-2:])
            ctp_nbr = int(ctp_id[-4:])
            assert_equal(ctp_nbr, qin_nbr, "Patient number incorrect; expected:"
                         " %d found: %d" % (qin_nbr, ctp_nbr))


if __name__ == "__main__":
    import nose
    nose.main(defaultTest=__name__)



class BolusArrivalError(Exception):
    pass


def bolus_arrival_index(time_series):
    """
    Determines the DCE bolus arrival series index. The bolus arrival is
    the first series with a difference in average signal larger than
    double the difference from first two points.
 
    :param time_series: the 4D NiFTI scan image file path
    :return: the bolus arrival series index
    :raise BolusArrivalError: if the bolus arrival could not be determined
    """
    import nibabel as nb
    import numpy as np

    nii = nb.load(time_series)
    data = nii.get_data()
    n_vols = data.shape[-1]
    signal_means = np.array([np.mean(data[:,:,:, idx])
                             for idx in xrange(n_vols)])
    signal_diffs = np.diff(signal_means)

    # If we see a difference in average signal larger than double the
    # difference from first two points, take that as bolus arrival.
    base_diff = np.abs(signal_diffs[0])
    for idx, diff_val in enumerate(signal_diffs[1:]):
        if diff_val > 2 * base_diff:
            return idx + 1
    else:
        raise BolusArrivalError("Unable to determine bolus arrival")



import os
import re
import tempfile
import logging
from collections import defaultdict
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import (IdentityInterface, Function, Merge)
from nipype.interfaces.dcmstack import MergeNifti
import qipipe
from . import staging
from .workflow_base import WorkflowBase
from .staging import StagingWorkflow
from .mask import MaskWorkflow
import registration
from ..interfaces import (XNATDownload, XNATUpload)
import qixnat
from qiutil.logging import logger
from ..staging.staging_helper import iter_stage
from ..staging.map_ctp import map_ctp

SCAN_TS_RSC = 'scan_ts'
"""The XNAT scan time series resouce name."""

MASK_RSC = 'mask'
"""The XNAT mask resouce name."""


def run(*inputs, **opts):
    """
    Creates a :class:`qipipe.pipeline.qipipeline.QIPipelineWorkflow`
    and runs it on the given inputs.
    
    If the *actions* option includes ``stage``, then the input is
    the :meth:`QIPipelineWorkflow.run_with_dicom_input` DICOM
    directories input. Otherwise, the input is the
    :meth:`QIPipelineWorkflow.run_with_scan_download` XNAT session
    labels input.

    :param inputs: the DICOM directories or XNAT session labels to
        process
    :param opts: the :meth:`qipipe.staging.staging_helper.iter_stage`
        and :class:`QIPipelineWorkflow` initializer options,
        as well as the following keyword options:
    :keyword collection: the AIRC collection name
    :keyword resume: flag indicating whether to resume processing on
        existing sessions (default False)
    """
    # Tailor the actions.
    actions = opts.get('actions', _default_actions(**opts))
    opts['actions'] = actions
    if 'stage' in actions:
        _run_with_dicom_input(*inputs, **opts)
    else:
        _run_with_xnat_input(*inputs, **opts)


def _default_actions(**opts):
    """
    Returns the default actions from the given options, as follows:

    * The default actions always include ``model``.

    * If the ``registration`` resource option is not set, then ``stage``
      is included in the actions.

    :param opts: the :meth:`run` options
    :return: the default actions
    """
    if 'registration' in opts:
        return ['register', 'model']
    else:
        return  ['stage', 'register', 'model']


def _run_with_dicom_input(*inputs, **opts):
    """
    :param inputs: the :meth:`QIPipelineWorkflow.run_with_dicom_input`
        inputs
    :param opts: the :meth:`run` options as well as the following keyword
        options:
    :keyword collection: the AIRC collection name
    :keyword resume: flag indicating whether to resume workflow execution
    """
    # The required AIRC collection.
    collection = opts.pop('collection', None)
    if not collection:
        raise ValueError('The staging pipeline collection is missing.')
    # The target directory.
    dest = opts.get('dest', None)
    # The resume option corresponds to the staging helper iter_stage
    # function skip_existing option.
    if opts.pop('resume', None):
        opts['skip_existing'] = False
        
    # The set of input subjects is used to build the CTP mapping file
    # after the workflow is completed.
    subjects = set()
    # Run the workflow on each session and scan type.
    for sbj, sess, scan_type_dict in iter_stage(collection, *inputs, **opts):
        # Capture the subject.
        subjects.add(sbj)
        # Pull out the actions, since the non-stage actions only apply
        # to the T1 scan type.
        actions = opts.pop('actions')
        # Run the workflow on each scan type.
        for scan_type, scan_dict in scan_type_dict.iteritems():
            opts['scan_type'] = scan_type
            # Only T1 can do more than staging.
            if scan_type == 't1':
                opts['actions'] = actions
            else:
                opts['actions'] = ['stage']
            # Create a new workflow for the current scan type.
            wf_gen = QIPipelineWorkflow(scan_type=scan_type)
            # Run the workflow on each {series: [DICOM files]} item.
            wf_gen.run_with_dicom_input(collection, sbj, sess,
                                        scan_dict, dest)
    
    # Make the TCIA subject map.
    map_ctp(collection, *subjects, dest=dest)


def _run_with_xnat_input(*inputs, **opts):
    """
    Run the pipeline with a XNAT download.
    Each input is either a session label or path, e.g.
    ``Breast012_Session03`` or  ``Breast012/Session03``.
    
    :param inputs: the XNAT session labels or paths
    :param opts: the ``project`` and :class:`QIPipelineWorkflow`
        initializer options
    """
    prj = opts.pop('project', qipipe.project())
    with qixnat.connect() as xnat:
        for label in inputs:
            # Convert a path to a label.
            if '/' in label:
                label = label.sub('/', '_')
            sbj, sess = qixnat.parse_session_label(label)
            # Check for an existing mask.
            mask_obj = xnat.find(project=prj, subject=sbj, session=sess,
                                 resource=MASK_RSC)
            if mask_obj and mask_obj.files().get():
                opts['mask'] = MASK_RSC
                status = 'found'
            else:
                status = 'not found'
            logger(__name__).debug("The %s %s resource %s was %s." %
                                (sbj, sess, MASK_RSC, status))
            
            # If registration or modeling will be performed, then check
            # for an existing scan time series.
            if 'register' in opts['actions'] or 'model' in opts['actions']:
                scan_ts_obj = xnat.find(project=prj, subject=sbj, session=sess,
                                        resource=SCAN_TS_RSC)
                if scan_ts_obj and scan_ts_obj.files().get():
                    opts['scan_time_series'] = SCAN_TS_RSC
                    status = 'found'
                else:
                    status = 'not found'
                logger(__name__).debug("The %s %s scan time series resource %s was"
                                   " %s." % (sbj, sess, SCAN_TS_RSC, status))
            
            # If modeling will be performed on a specified registration
            # resource, then check for an existing realigned time series.
            if 'model' in opts['actions'] and 'registration' in opts:
                reg_ts_rsc = opts['registration'] + '_ts'
                reg_ts_obj = xnat.find(project=prj, subject=sbj, session=sess,
                             resource=reg_ts_rsc)
                if reg_ts_obj and reg_ts_obj.files().get():
                    opts['realigned_time_series'] = reg_ts_rsc
                    status = 'found'
                else:
                    status = 'not found'
                logger(__name__).debug("The %s %s realigned time series resource %s"
                                   " was %s." % (sbj, sess, reg_ts_rsc, status))

            # Execute the workflow.
            wf_gen = QIPipelineWorkflow(**opts)
            wf_gen.run_with_scan_download(xnat, prj, sbj, sess)


class ArgumentError(Exception):
    pass


class NotFoundError(Exception):
    pass


class QIPipelineWorkflow(WorkflowBase):
    """
    QIPipeline builds and executes the OHSU QIN workflows.
    The pipeline builds a composite workflow which stitches together
    the following constituent workflows:

    - staging: Prepare the new AIRC DICOM visits, as described in
      :class:`qipipe.pipeline.staging.StagingWorkflow`

    - mask: Create the mask from the staged images,
      as described in :class:`qipipe.pipeline.mask.MaskWorkflow`

    - registration: Mask, register and realign the staged images,
      as described in
      :class:`qipipe.pipeline.registration.RegistrationWorkflow`

    - modeling: Perform PK modeling as described in
      :class:`qipipe.pipeline.modeling.ModelingWorkflow`

    The constituent workflows are determined by the initialization
    options ``stage``, ``register`` and ``model``. The default is
    to perform each of these subworkflows.
    
    The workflow steps are determined by the input options as follows:
    
    - If staging is performed, then the DICOM files are staged for the
      subject directory inputs. Otherwise, staging is not performed.
      In that case, if registration is enabled as described below, then
      the previously staged series scan stack images are downloaded.

    - If registration is performed and the ``registration`` resource option
      is set, then the previously realigned images with the given resource
      name are downloaded. The remaining scans are registered.

    - If registration or modeling is performed and the XNAT ``mask``
      resource is found, then that resource file is downloaded. Otherwise,
      the mask is created from the staged images.

    The QIN workflow input node is *input_spec* with the following
    fields:

    - *subject*: the subject name

    - *session*: the session name

    In addition, if the staging or registration workflow is enabled
    then the *iter_series* node iterables input includes the
    following fields:

    - *series*: the scan number

    - *dest*: the target staging directory, if the staging
      workflow is enabled

    The constituent workflows are combined as follows:

    - The staging workflow input is the QIN workflow input.

    - The mask and reference workflow input images is the newly or
      previously staged scan NiFTI image files.

    - The modeling workflow input images is the combination of previously
      and newly realigned image files.

    The pipeline execution workflow is also available as the *workflow*
    instance variable. The workflow input node is named *input_spec*
    with the same input fields as the
    :class:`qipipe.staging.RegistrationWorkflow` workflow *input_spec*.
    """

    REG_SERIES_PAT = re.compile('series(\d+)_reg_')

    def __init__(self, **opts):
        """
        :param opts: the :class:`qipipe.staging.WorkflowBase`
            initialization options as well as the following options:
        :keyword base_dir: the workflow execution directory
            (default a new temp directory)
        :keyword mask: the XNAT mask reconstruction name
        :keyword registration: the XNAT registration reconstruction name
        :keyword technique: the
            class:`qipipe.pipeline.registration.RegistrationWorkflow`
            technique
        """
        # The dry run and configuration file options are processed by methods
        # defined in the superclass WorkflowBase. The remaining options are
        # processed by methods defined in this QIPipelineWorkflow class.
        base_opts = {k: opts.pop(k) for k in ['cfg_file', 'dry_run'] if k in opts}
        super(QIPipelineWorkflow, self).__init__(logger(__name__), **base_opts)

        self.registration_resource = None
        """The registration XNAT reconstruction name."""

        self.modeling_resource = None
        """The modeling XNAT resource name."""

        self.workflow = self._create_workflow(**opts)
        """
        The pipeline execution workflow. The execution workflow is executed by
        calling the :meth:`run_with_dicom_input` or
        :meth:`run_with_scan_download` method.
        """

    def run_with_dicom_input(self, collection, subject, session, scan_dict,
                             dest=None):
        """
        :param collection: the AIRC collection name
        :param subject: the subject name
        :param session: the session name
        :param scan_dict: the *{scan number: [DICOM files]}* dictionary
        :param dest: the TCIA staging destination directory (default is
            the current working directory)
        """
        staging.set_workflow_inputs(self.workflow, collection, subject,
                                    session, scan_dict, dest)
        self._run_workflow(self.workflow)

    def run_with_scan_download(self, xnat, project, subject, session):
        """
        Runs the execution workflow on downloaded scan image files.
        """
        self._logger.debug("Processing the %s %s %s scans..." %
                           (project, subject, session))

        # Get the scan numbers.
        scans = xnat.get_scan_numbers(project, subject, session)
        if not scans:
            raise NotFoundError("The QIN pipeline did not find a %s %s %s"
                                " scan." % (project, subject, session))

        # Partition the input scans into those which are already
        # registered and those which need to be registered.
        reg_scans, unreg_scans = self._partition_scans(xnat, project, subject,
                                                       session, scans)

        # Set the workflow input.
        input_spec = self.workflow.get_node('input_spec')
        input_spec.inputs.subject = subject
        input_spec.inputs.session = session
        input_spec.inputs.unregistered_series = unreg_scans

        # Execute the workflow.
        self._run_workflow(self.workflow)
        self._logger.debug("Processed %d %s %s %s scans." %
                           (len(scans), project, subject, session))

    def _partition_scans(self, xnat, project, subject, session, scans):
        """
        Partitions the given scans into those which have a corresponding
        registration reconstruction file and those which don't.

        :return: the (registered, unregistered) scan numbers
        """
        # The XNAT registration object.
        if self.registration_resource:
            reg_obj = xnat.get_experiment_resource(project, subject, session,
                                                   self.registration_resource)
        else:
            reg_obj = None
        # If the registration has not yet been performed, then
        # download all of the scans.
        if not (reg_obj and reg_obj.exists()):
            return [], scans

        # The realigned scan numbers.
        reg_scans = set(self._registered_scans(reg_obj))
        
        # The unregistered scan numbers.
        unreg_scans = set(scans) - reg_scans
        self._logger.debug("The %s %s %s resource has %d registered scans"
                           " and %d unregistered scans." %
                           (project, subject, session, len(reg_scans),
                            len(unreg_scans)))

        return (reg_scans, unreg_scans)

    def _registered_scans(self, reg_obj):
        """
        Returns the scans which have a corresponding registration
        reconstruction file.

        :param reg_obj: the XNAT registration reconstruction object
        :return: the registered scan numbers
        """
        # The XNAT registration file names.
        reg_files = reg_obj.files().get()
        # Match on the realigned scan file pattern.
        matches = ((QIPipelineWorkflow.REG_SERIES_PAT.match(f)
                    for f in reg_files))

        return [int(match.group(1)) for match in matches if match]

    def _create_workflow(self, **opts):
        """
        Builds the reusable pipeline workflow described in
        :class:`qipipe.pipeline.qipipeline.QIPipeline`.

        :param opts: the constituent workflow initializer options
        :return: the Nipype workflow
        """
        self._logger.debug("Building the QIN pipeline execution workflow...")

        # This is a long method body with the following stages:
        #
        # 1. Gather the options.
        # 2. Create the constituent workflows.
        # 3. Tie together the constituent workflows.
        #
        # The constituent workflows are created in back-to-front order,
        # i.e. modeling, registration, reference, mask, staging.
        # This order makes it easier to determine whether to create
        # an upstream workflow depending on the presence of downstream
        # workflows, e.g. the mask is not created if registration
        # is not performed.
        #
        # By contrast, the workflows are tied together in front-to-back
        # order.

        # The work directory used for the master workflow and all
        # constituent workflows.
        if 'base_dir' in opts:
            base_dir = os.path.abspath(opts['base_dir'])
        else:
            base_dir = tempfile.mkdtemp()

        # The execution workflow.
        exec_wf = pe.Workflow(name='qipipeline', base_dir=base_dir)

        # The workflow options.
        actions = opts.get('actions', ['stage', 'register', 'model'])
        mask_rsc = opts.get('mask')
        scan_ts_rsc = opts.get('scan_time_series')
        reg_rsc = opts.get('registration')
        reg_ts_rsc = opts.get('realigned_time_series')
        reg_technique = opts.get('technique')

        # Set the project, if necessary.
        prj = opts.pop('project', None)
        if prj:
            qipipe.project(prj)
            self._logger.debug("Set the XNAT project to %s." % prj)
        
        # Set the registration resource instance variable.
        if reg_rsc:
            self.registration_resource = reg_rsc

        # The modeling workflow. Since the proprietary fastfit module might
        # not be available, only import the ModelingWorkflow on demand if
        # modeling is required.
        if 'model' in actions:
            from .modeling import ModelingWorkflow
            mdl_wf_gen = ModelingWorkflow(base_dir=base_dir)
            mdl_wf = mdl_wf_gen.workflow
            self.modeling_resource = mdl_wf_gen.resource
        else:
            mdl_wf = None

        # The registration workflow node.
        if 'register' in actions:
            reg_inputs = ['subject', 'session', 'in_files',
                          'bolus_arrival_index', 'mask', 'opts']
            
            # The registration function keyword options.
            reg_opts = dict(base_dir=base_dir)
            if reg_technique:
                reg_opts['technique'] = reg_technique
            # If the resource was not specified, then make a new
            # resource name.
            if not reg_rsc:
                new_reg_rsc = registration.generate_resource_name()
                self.registration_resource = new_reg_rsc
            # Add the resource name to the registration options.
            reg_opts['resource'] = self.registration_resource
            
            # The registration function.
            reg_xfc = Function(input_names=reg_inputs,
                               output_names=['out_files'],
                               function=register_scans)
            reg_node = pe.Node(reg_xfc, name='registration')
            reg_node.inputs.opts = reg_opts
        else:
            self._logger.info("Skipping registration.")
            reg_node = None

        # Registration and modeling require a mask.
        if (reg_node or mdl_wf) and not mask_rsc:
            mask_wf = MaskWorkflow(base_dir=base_dir).workflow
        else:
            self._logger.info("Skipping mask creation.")
            mask_wf = None

        # The staging workflow.
        if 'stage' in actions:
            scan_type = opts.pop('scan_type', None)
            if not scan_type:
                raise ArgumentError("The required staging argument scan_type is missing")
            stg_wf = StagingWorkflow(scan_type, base_dir=base_dir).workflow
        else:
            self._logger.info("Skipping staging.")
            stg_wf = None

        # Validate that there is at least one constituent workflow.
        if not any([stg_wf, reg_node, mdl_wf]):
            raise ArgumentError("No workflow was enabled.")

        # The workflow input fields.
        input_fields = ['subject', 'session']
        iter_series_fields = ['series']
        # The staging workflow has additional input fields.
        # Partial registration requires the unregistered scans input.
        if stg_wf:
            input_fields.append('collection')
            iter_series_fields.append('dest')
        elif reg_node and reg_rsc:
            input_fields.append('unregistered_series')
        
        # The workflow input node.
        input_spec_xfc = IdentityInterface(fields=input_fields)
        input_spec = pe.Node(input_spec_xfc, name='input_spec')
        # Most workflows require a series iterator node.
        if stg_wf or reg_node or mask_wf or (mdl_wf and not scan_ts_rsc):
            iter_series_xfc = IdentityInterface(fields=iter_series_fields)
            iter_series = pe.Node(iter_series_xfc, name='iter_series')
        # Staging requires a DICOM iterator node.
        if stg_wf:
            iter_dicom_xfc = IdentityInterface(fields=['series', 'dicom_file'])
            iter_dicom = pe.Node(iter_dicom_xfc, name='iter_dicom')
            exec_wf.connect(iter_series, 'series', iter_dicom, 'series')

        # Stitch together the workflows:

        # If staging is enabled, then stage the DICOM input.
        # Otherwise, if the registration, mask or reference
        # workflow is enabled, then download the scan files.
        if stg_wf:
            for field in input_spec.inputs.copyable_trait_names():
                exec_wf.connect(input_spec, field,
                                stg_wf, 'input_spec.' + field)
            for field in iter_series.inputs.copyable_trait_names():
                exec_wf.connect(iter_series, field,
                                stg_wf, 'iter_series.' + field)
            exec_wf.connect(iter_dicom, 'dicom_file',
                            stg_wf, 'iter_dicom.dicom_file')

        # Some workflows require the scans. If staging is
        # enabled then collect the staged NiFTI scan images. Otherwise,
        # download the XNAT NiFTI scan images. In either case, the
        # staged images are collected in a node named 'staged' with
        # output 'images'.
        if reg_node or mask_wf or (mdl_wf and not scan_ts_rsc):
            if stg_wf:
                staged = pe.JoinNode(IdentityInterface(fields=['out_files']),
                                    joinsource='iter_series', name='staged')
                exec_wf.connect(stg_wf, 'output_spec.image',
                                staged, 'out_files')
            else:
                dl_scans_xfc = XNATDownload(project=qipipe.project(),
                                            container_type='scan',
                                            resource='NIFTI')
                staged = pe.Node(dl_scans_xfc, name='staged')
                exec_wf.connect(input_spec, 'subject', staged, 'subject')
                exec_wf.connect(input_spec, 'session', staged, 'session')

        # Registration and modeling require a time series, mask and
        # bolus arrival.
        if reg_node or mdl_wf:
            # If a time series resource name was specified, then download
            # the time series. Otherwise, make the time series.
            if scan_ts_rsc:
                dl_scan_ts_xfc = XNATDownload(project=qipipe.project(),
                                              resource=scan_ts_rsc)
                scan_ts = pe.Node(dl_scan_ts_xfc,
                                  name='download_scan_time_series')
                exec_wf.connect(input_spec, 'subject', scan_ts, 'subject')
                exec_wf.connect(input_spec, 'session', scan_ts, 'session')
            else:
                # Merge the staged files.
                scan_ts_xfc = MergeNifti(out_format=SCAN_TS_RSC)
                scan_ts = pe.Node(scan_ts_xfc, name='merge_scans')
                exec_wf.connect(staged, 'out_files', scan_ts, 'in_files')
                # Upload the time series.
                ul_scan_ts_xfc = XNATUpload(project=qipipe.project(),
                                            resource=SCAN_TS_RSC)
                ul_scan_ts = pe.Node(ul_scan_ts_xfc,
                                     name='upload_scan_time_series')
                exec_wf.connect(input_spec, 'subject', ul_scan_ts, 'subject')
                exec_wf.connect(input_spec, 'session', ul_scan_ts, 'session')
                exec_wf.connect(scan_ts, 'out_file', ul_scan_ts, 'in_files')

            # If a mask resource name was specified, then download the mask.
            # Otherwise, make the mask.
            if mask_rsc:
                dl_mask_xfc = XNATDownload(project=qipipe.project(),
                                           resource=mask_rsc)
                download_mask = pe.Node(dl_mask_xfc, name='download_mask')
                exec_wf.connect(input_spec, 'subject', download_mask, 'subject')
                exec_wf.connect(input_spec, 'session', download_mask, 'session')
            else:
                assert mask_wf, "The mask workflow is missing"
                exec_wf.connect(input_spec, 'subject',
                                mask_wf, 'input_spec.subject')
                exec_wf.connect(input_spec, 'session',
                                mask_wf, 'input_spec.session')
                exec_wf.connect(scan_ts, 'out_file',
                                mask_wf, 'input_spec.time_series')
            
            # Compute the bolus arrival from the scan time series.
            bolus_arv_xfc = Function(input_names=['time_series'],
                                     output_names=['bolus_arrival_index'],
                                     function=bolus_arrival_index_or_zero)
            bolus_arv = pe.Node(bolus_arv_xfc, name='bolus_arrival_index')
            exec_wf.connect(scan_ts, 'out_file', bolus_arv, 'time_series')

        # If registration is enabled, then register the staged images.
        if reg_node:
            exec_wf.connect(input_spec, 'subject', reg_node, 'subject')
            exec_wf.connect(input_spec, 'session', reg_node, 'session')
            # The staged input.
            if stg_wf or not reg_rsc:
                exec_wf.connect(staged, 'out_files', reg_node, 'in_files')
            else:
                # Select only the unregistered scans.
                sel_unreg_xfc = Function(input_names=['scans', 'in_files'],
                                         output_names=['out_files'],
                                         function=select_scan_files)
                unreged = pe.Node(sel_unreg_xfc, name='unregistered')
                exec_wf.connect(input_spec, 'unregistered_series',
                                unreged, 'scans')
                exec_wf.connect(staged, 'out_files', unreged, 'in_files')
                exec_wf.connect(unreged, 'out_files', reg_node, 'in_files')
            # The mask input.
            if mask_wf:
                exec_wf.connect(mask_wf, 'output_spec.mask', reg_node, 'mask')
            else:
                exec_wf.connect(download_mask, 'out_file', reg_node, 'mask')
            # The bolus arrival.
            exec_wf.connect(bolus_arv, 'bolus_arrival_index',
                            reg_node, 'bolus_arrival_index')

        # If the modeling workflow is enabled, then model the scan or realigned
        # images.
        if mdl_wf:
            exec_wf.connect(input_spec, 'subject',
                            mdl_wf, 'input_spec.subject')
            exec_wf.connect(input_spec, 'session',
                            mdl_wf, 'input_spec.session')
            # The mask input.
            if mask_wf:
                exec_wf.connect(mask_wf, 'output_spec.mask',
                                mdl_wf, 'input_spec.mask')
            else:
                exec_wf.connect(download_mask, 'out_file',
                                mdl_wf, 'input_spec.mask')
            # The bolus arrival.
            exec_wf.connect(bolus_arv, 'bolus_arrival_index',
                            mdl_wf, 'input_spec.bolus_arrival_index')

            # If registration is enabled, then the 4D time series
            # is created by that workflow, otherwise download the
            # previously created 4D time series.
            if reg_ts_rsc:
                # Download the XNAT time series file.
                ts_dl_xfc = XNATDownload(project=qipipe.project(),
                                         resource=reg_ts_rsc)
                reg_ts = pe.Node(ts_dl_xfc, name='download_reg_time_series')
                exec_wf.connect(input_spec, 'subject', reg_ts, 'subject')
                exec_wf.connect(input_spec, 'session', reg_ts, 'session')
                exec_wf.connect(reg_ts, 'out_file',
                                mdl_wf, 'input_spec.time_series')
            elif self.registration_resource:
                # Merge the realigned images to 4D.
                reg_ts_rsc = self.registration_resource + '_ts'
                merge_reg = pe.Node(MergeNifti(out_format=reg_ts_rsc),
                                    name='merge_reg')
                
                # If the registration reconstruction name was specified,
                # then download the previously realigned images.
                if reg_rsc:
                    reg_dl_xfc = XNATDownload(project=qipipe.project(),
                                              resource=reg_rsc)
                    download_reg = pe.Node(reg_dl_xfc,
                                           name='download_realigned_images')
                    exec_wf.connect(input_spec, 'subject',
                                    download_reg, 'subject')
                    exec_wf.connect(input_spec, 'session',
                                    download_reg, 'session')
                    if reg_node:
                        # Merge the previously and newly realigned images.
                        concat_reg = pe.Node(Merge(2), name='concat_reg')
                        exec_wf.connect(download_reg, 'out_files',
                                        concat_reg, 'in1')
                        exec_wf.connect(reg_node, 'out_files',
                                        concat_reg, 'in2')
                        exec_wf.connect(concat_reg, 'out',
                                        merge_reg, 'in_files')
                    else:
                        # All of the realigned files were downloaded.
                        exec_wf.connect(download_reg, 'out_files',
                                        merge_reg, 'in_files')
                elif reg_node:
                    # All of the realigned files were created by the
                    # registration workflow.
                    exec_wf.connect(reg_node, 'out_files',
                                    merge_reg, 'in_files')
                else:
                    raise ArgumentError(
                        "The QIN pipeline cannot perform modeling on the"
                        " registration result, since the registration"
                        " workflow is disabled and no registration resource"
                        " was specified.")

                # Upload the realigned time series to XNAT.
                upload_reg_ts_xfc = XNATUpload(project=qipipe.project(),
                                               resource=reg_ts_rsc)
                upload_reg_ts = pe.Node(upload_reg_ts_xfc,
                                        name='upload_reg_time_series')
                exec_wf.connect(input_spec, 'subject',
                                upload_reg_ts, 'subject')
                exec_wf.connect(input_spec, 'session',
                                upload_reg_ts, 'session')
                exec_wf.connect(merge_reg, 'out_file',
                                upload_reg_ts, 'in_files')
                
                # Pass the realigned time series to modeling.
                exec_wf.connect(merge_reg, 'out_file',
                                mdl_wf, 'input_spec.time_series')
            else:
                # Model the scan input.
                exec_wf.connect(scan_ts, 'out_file',
                                mdl_wf, 'input_spec.time_series')

        # Set the configured workflow node inputs and plug-in options.
        self._configure_nodes(exec_wf)

        self._logger.debug("Created the %s workflow." % exec_wf.name)
        # If debug is set, then diagram the workflow graph.
        if self._logger.level <= logging.DEBUG:
            self.depict_workflow(exec_wf)

        return exec_wf

    def _run_workflow(self, workflow):
        """
        Overrides the superclass method to build the workflow if the
        *dry_run* instance variable flag is set.

        :param workflow: the workflow to run
        """
        super(QIPipelineWorkflow, self)._run_workflow(workflow)
        if self.dry_run:
            _, path = tempfile.mkstemp()
            try:
                opts = dict(base_dir=self.workflow.base_dir, dry_run=True)
                register_scans('Dummy', 'Dummy', [path], 0, path, opts)
            finally:
                os.remove(path)


def bolus_arrival_index_or_zero(time_series):
    from qipipe.helpers.bolus_arrival import (bolus_arrival_index,
                                              BolusArrivalError)

    # Determines the bolus uptake. If it could not be determined,
    # then the first series is taken to be the uptake.
    try:
        return bolus_arrival_index(time_series)
    except BolusArrivalError:
        return 0


def select_scan_files(scans, in_files):
    """
    :param scans: the scan numbers
    :param in_files: the scan files
    :return: the scan files for the given series 
    """
    import re
    
    scans = set(scans)
    series_pat = re.compile("series(\d{3}).nii.gz$")
    return [f for f in in_files
            if int(series_pat.search(f).group(1)) in scans]


def register_scans(subject, session, in_files, bolus_arrival_index,
                   mask, opts):
    """
    Runs the registration workflow on the given session scan images.

    :Note: contrary to Python convention, the opts method parameter
      is a required dictionary rather than a keyword aggregate (e.g.
      ``**opts``). The Nipype ``Function`` interface does not support
      method aggregates.

    :param subject: the subject name
    :param session: the session name
    :param in_files: the input session scan images
    :param mask: the image mask file path
    :param bolus_arrival_index: the bolus uptake series index
    :param opts: the :meth:`qipipe.pipeline.registration.run` keyword
        options
    :return: the realigned image file path array
    """
    from qipipe.pipeline import registration

    return registration.run(subject, session, in_files, bolus_arrival_index,
                            mask, **opts)



import os
import re
from qidicom import meta
from .sarcoma_config import sarcoma_location

from qiutil.logging import logger


def fix_dicom_headers(collection, subject, *dicom_files, **opts):
    """
    Fix the given input OHSU QIN AIRC DICOM files as follows:

    - Replace the ``Patient ID`` value with the subject number, e.g.
        ``Sarcoma001``

    - Add the ``Body Part Examined`` tag

    - Standardize the file name

    The ``Body Part Examined`` tag is set as follows:

    - If the collection is ``Sarcoma``, then the body part is the
        :meth:`qipipe.staging.sarcoma_config.sarcoma_location`.
    
    - Otherwise, the body part is the capitalized collection name, e.g.
        ``BREAST``.
        
    The output file name is standardized as follows:

    - The file name is lower-case

    - The file extension is ``.dcm``
    
    - The scan type is appended to the fie name, e.g. ``_t1``

    - Each non-word character is replaced by an underscore
    
    :param collection: the collection name
    :param subject: the input subject name
    :param opts: the following keyword arguments:
    :keyword dest: the location in which to write the modified files
        (default is the current directory)
    :return: the files which were created
    :raise StagingError: if the collection is not supported
    """

    # Make the tag name => value dictionary.
    if collection == 'Sarcoma':
        site = sarcoma_location(subject)
    else:
        site = collection.upper()
    tnv = dict(PatientID=subject, BodyPartExamined=site)

    # Set the tags in each image file.
    if 'dest' in opts:
        dest = opts['dest']
    else:
        dest = os.getcwd()
    edited = meta.edit(dest, *dicom_files, **tnv)

    # Rename the edited files as necessary.
    out_files = []
    for f in edited:
        std_name = _standardize_dicom_file_name(f)
        if f != std_name:
            os.rename(f, std_name)
            out_files.append(std_name)
        logger(__name__).debug(
            "The DICOM headers in %s were fixed and saved as %s." % (f, std_name))

    return out_files


def _standardize_dicom_file_name(path):
    """
    Standardizes the given input file name.
    """
    fdir, fname = os.path.split(path)
    # Replace non-word characters.
    fname = re.sub('\W', '_', fname.lower())
    # Add a .dcm extension, if necessary.
    _, ext = os.path.splitext(fname)
    if not ext:
        fname = fname + '.dcm'
    return os.path.join(fdir, fname)



import tempfile
import logging
from nipype.pipeline import engine as pe
from nipype.interfaces import fsl
from nipype.interfaces.dcmstack import CopyMeta
from nipype.interfaces.utility import (IdentityInterface, Function, Merge)
import qiutil
from qiutil.logging import logger
from qixnat import project
from ..interfaces import XNATUpload #, UpdateQIProfile)
from ..helpers.bolus_arrival import bolus_arrival_index, BolusArrivalError
from .workflow_base import WorkflowBase
from .distributable import DISTRIBUTABLE


PK_PREFIX = 'pk'
"""The XNAT modeling resource object label prefix."""


def run(subject, session, *images, **opts):
    """
    Creates a :class:`qipipe.pipeline.modeling.ModelingWorkflow` and
    runs it on the given inputs.

    :param subject: the input subject
    :param session: the input session
    :param images: the input 3D NiFTI images
    :param opts: the :class:`qipipe.pipeline.modeling.ModelingWorkflow`
        initializer options
    :return: the :meth:`qipipe.pipeline.modeling.ModelingWorkflow.run`
        result
    """
    return ModelingWorkflow(**opts).run(subject, session, *images)


class ModelingWorkflow(WorkflowBase):

    """
    The ModelingWorkflow builds and executes the Nipype pharmacokinetic
    mapping workflow.

    The workflow calculates the modeling parameters for an input 4D
    time series NiFTI image file as follows:

    - Compute the |R10| value, if it is not given in the options

    - Convert the DCE time series to a R1 map series

    - Determine the AIF and R1 fit parameters from the time series

    - Optimize the BOLERO pharmacokinetic model

    - Upload the modeling result to XNAT

    The modeling workflow input is the `input_spec` node consisting of the
    following input fields:

    - *subject*: the subject name

    - *session*: the session name

    - *mask*: the mask to apply to the images

    - *time_series*: the 4D time series NiFTI file to model

    - *bolus_arrival_index*: the bolus uptake series index

    - the PK modeling parameters described below

    If an input field is defined in the configuration file ``Parameters``
    topic, then the input field is set to that value.

    If the |R10| option is not set, then it is computed from the proton
    density weighted scans and DCE series baseline image.

    The outputs are collected in the `output_spec` node for the FXL
    (`Tofts standard`_) model and the FXR (`shutter speed`_) model with
    the following fields:

    - `r1_series`: the R1 series files

    - `pk_params`: the AIF and R1 parameter CSV file

    - `fxr_k_trans`, `fxl_k_trans`: the |Ktrans| vascular permeability
       transfer constant

    - `delta_k_trans`: the FXR-FXL |Ktrans| difference

    - `fxr_v_e`, `fxl_v_e`: the |ve| extravascular extracellular volume
       fraction

    - `fxr_tau_i`: the intracellular |H2O| mean lifetime

    - `fxr_chisq`, `fxl_chisq`: the intracellular |H2O| mean lifetime

    In addition, if |R10| is computed, then the output includes the
    following fields:

    - `pdw_image`: the proton density weighted image

    - `dce_baseline`: the DCE series baseline image

    - `r1_0`: the computed |R10| value

    This workflow is adapted from the `AIRC DCE`_ implementation.

    :Note: This workflow uses proprietary OHSU AIRC software, notably the
        BOLERO implementation of the shutter speed model.

    .. reST substitutions:
    .. |H2O| replace:: H\ :sub:`2`\ O
    .. |Ktrans| replace:: K\ :sup:`trans`
    .. |ve| replace:: v\ :sub:`e`
    .. |R10| replace:: R1\ :sub:`0`

    .. _Tofts standard: http://onlinelibrary.wiley.com/doi/10.1002/(SICI)1522-2586(199909)10:3%3C223::AID-JMRI2%3E3.0.CO;2-S/abstract
    .. _shutter speed: http://www.ncbi.nlm.nih.gov/pmc/articles/PMC2582583
    .. _AIRC DCE: https://everett.ohsu.edu/hg/qin_dce
    """

    def __init__(self, **opts):
        """
        The modeling parameters can be defined in either the options or the
        configuration as follows:

        - The parameters can be defined in the configuration
          ``Parameters`` section.

        - The input options take precedence over the configuration
          settings.

        - The *r1_0_val* takes precedence over the R1_0 computation
          fields *pd_dir* and *max_r1_0*. If *r1_0_val* is set
          in the input options, then *pd_dir* and *max_r1_0* are
          not included from the result.

        - If *pd_dir* and *max_r1_0* are set in the input options
          and *r1_0_val* is not set in the input options, then
          a *r1_0_val* configuration setting is ignored.

        - The *baseline_end_idx* defaults to 1 if it is not set in
          either the input options or the configuration.

        :param opts: the following initialization options:
        :keyword cfg_file: the optional workflow inputs configuration file
        :keyword base_dir: the workflow execution directory
            (default a new temp directory)
        :keyword r1_0_val: the optional fixed |R10| value
        :keyword max_r1_0: the maximum computed |R10| value, if the fixed
            |R10| option is not set
        :keyword pd_dir: the proton density files parent directory, if the
            fixed |R10| option is not set
        :keyword baseline_end_idx: the number of series to merge into a R1
            series baseline image (default is 1)
        """
        cfg_file = opts.pop('cfg_file', None)
        super(ModelingWorkflow, self).__init__(logger(__name__), cfg_file)

        resource = opts.pop('modeling', None)
        if not resource:
            resource = self._generate_resource_name()
        self.resource = resource
        """
        The XNAT resource name for all executions of this
        :class:`qipipe.pipeline.modeling.ModelingWorkflow` instance. The
        name is unique, which permits more than one model to be stored for
        each input series without a name conflict.
        """

        self.workflow = self._create_workflow(**opts)
        """
        The modeling workflow described in
        :class:`qipipe.pipeline.modeling.ModelingWorkflow`.
        """

    def run(self, input_dict, **opts):
        """
        Executes the modeling workflow described in
        :class:`qipipe.pipeline.modeling.ModelingWorkflow`
        on the given inputs.

        This run method connects the given inputs to the modeling workflow
        inputs. The execution workflow is then executed, resulting in a
        new uploaded XNAT analysis resource for each input session. This
        method returns the uploaded XNAT *(subject, session, analysis)*
        name tuples.

        The input is a *{subject: {session: image_dict}}* dictionary, where
        *image_dict* is a dictionary with the following content:

            {``time_series``: 4D time series, ``mask``: mask file}

        :param input_dict: the input *{subject: {session: image_dict}}*
            to model
        :param opts: the following workflow options:
        :keyword resource: the XNAT resource to model
        :return: the modeling result XNAT resource name
        """
        # The image dictionaries lists.
        img_dict_lists = [img_dict.values()
                     for img_dict in input_dict.itervalues()]
        # The number of sessions.
        sess_cnt = len(img_dict_lists)

        # Model the images.
        self._logger.debug("Modeling %d sessions..." % sess_cnt)
        for sbj, sess_dict in input_dict.iteritems():
            self._logger.debug("Modeling subject %s..." % sbj)
            for sess, img_dict in sess_dict.iteritems():
                self._logger.debug("Modeling %s %s..." % (sbj, sess))
                self._model(sbj, sess, **img_dict)
                self._logger.debug("Modeled %s %s." % (sbj, sess))
            self._logger.debug("Modeled subject %s." % sbj)
        self._logger.debug("Modeled %d images from %d sessions." %
                           (img_cnt, sess_cnt))

        # Return the analysis name.
        return self.resource

    def _generate_resource_name(self):
        """
        Makes a unique modeling resource name. Uniqueness permits more than
        one resource to be stored for a given session without a name conflict.

        :return: a unique XNAT modeling resource name
        """
        return "%s_%s" % (PK_PREFIX, qiutil.file.generate_file_name())

    def _model(self, subject, session, time_series, mask):
        """
        Runs the modeling workflow on the given session images.

        :param subject: the subject name
        :param session: the session name
        :param time_series: the 4D NiFTI time series image file
        :param mask: the mask file
        """
        self._logger.debug("Modeling the %s %s %s time series..." %
            (sbj, sess, time_series))

        # Determine the bolus uptake. If it could not be determined,
        # then take the first series as the uptake.
        try:
            bolus_arv_ndx = bolus_arrival_index(time_series)
        except BolusArrivalError:
            bolus_arv_ndx = 0

        # Set the workflow input.
        input_spec = self.workflow.get_node('input_spec')
        input_spec.inputs.subject = subject
        input_spec.inputs.session = session
        input_spec.inputs.time_series = time_series
        input_spec.inputs.bolus_arrival_index = bolus_arv_ndx
        input_spec.inputs.mask = mask

        # Execute the modeling workflow.
        self._logger.debug("Executing the %s workflow on the %s %s %s time"
                           " series..." %
                           (self.workflow.name, subject, session, time_series))
        self._run_workflow(self.workflow)
        self._logger.debug("Executed the %s workflow on the %s %s %s time"
                            " series." %
                           (self.workflow.name, subject, session, time_series))

    def _create_workflow(self, base_dir=None, **opts):
        """
        Builds the modeling workflow.

        :param base_dir: the execution working directory
            (default is a new temp directory)
        :param opts: the additional workflow initialization parameters
        :return: the Nipype workflow
        """
        self._logger.debug("Building the modeling workflow...")

        # The base workflow.
        if not base_dir:
            base_dir = tempfile.mkdtemp()
        mdl_wf = pe.Workflow(name='modeling', base_dir=base_dir)

        # Start with a base workflow.
        base_wf = self._create_base_workflow(base_dir=base_dir, **opts)

        # The workflow input fields.
        in_fields = ['subject', 'session', 'time_series', 'mask',
                     'bolus_arrival_index']
        input_xfc = IdentityInterface(fields=in_fields)
        input_spec = pe.Node(input_xfc, name='input_spec')
        self._logger.debug("The modeling workflow input is %s with"
            " fields %s" % (input_spec.name, in_fields))
        mdl_wf.connect(input_spec, 'time_series',
                       base_wf, 'input_spec.time_series')
        mdl_wf.connect(input_spec, 'mask', base_wf, 'input_spec.mask')
        mdl_wf.connect(input_spec, 'bolus_arrival_index',
                       base_wf, 'input_spec.bolus_arrival_index')

        # Upload the modeling results to XNAT.
        # Each output field contains a modeling result file.
        # Upload these files to the modeling resource.
        base_output = base_wf.get_node('output_spec')
        out_fields = base_output.outputs.copyable_trait_names()
        merge_outputs = pe.Node(Merge(len(out_fields)), name='merge_outputs')
        for i, field in enumerate(out_fields):
            base_field = 'output_spec.' + field
            mdl_wf.connect(base_wf, base_field, merge_outputs, "in%d" % (i + 1))
        upload_xfc = XNATUpload(project=project(), resource=self.resource)
        upload_node = pe.Node(upload_xfc, name='upload_outputs')
        mdl_wf.connect(input_spec, 'subject', upload_node, 'subject')
        mdl_wf.connect(input_spec, 'session', upload_node, 'session')
        mdl_wf.connect(merge_outputs, 'out', upload_node, 'in_files')

        # TODO - Get the FSL mean parameters.

        # TODO - Add the session to qiprofile.
        # update_profile_db = pe.Node(UpdateQIProfile())
        # mdl_wf.connect(input_spec, 'subject', upload_node, 'subject')
        # mdl_wf.connect(input_spec, 'session', upload_node, 'session')

        # The output is the modeling outputs.
        output_xfc = IdentityInterface(fields=out_fields)
        output_spec = pe.Node(output_xfc, name='output_spec')
        for field in out_fields:
            base_field = 'output_spec.' + field
            mdl_wf.connect(base_wf, base_field, output_spec, field)
        self._logger.debug("The modeling workflow output is %s with"
                           " fields %s" % (output_spec.name, out_fields))

        self._configure_nodes(mdl_wf)

        self._logger.debug("Created the %s workflow." % mdl_wf.name)
        # If debug is set, then diagram the workflow graph.
        if self._logger.level <= logging.DEBUG:
            self.depict_workflow(mdl_wf)

        return mdl_wf

    def _create_upload_node(self, resource):
        """
        :param resource: the modeling parameter resource name
        :return: the modeling parameter XNAT upload node
        """
        upload_xfc = XNATUpload(project=project(), resource=resource)
        name = 'upload_' + resource

        return pe.Node(upload_xfc, name=name)

    def _create_base_workflow(self, base_dir=None, **opts):
        """
        Creates the modeling base workflow. This workflow performs the steps
        described in :class:`qipipe.pipeline.modeling.ModelingWorkflow` with
        the exception of XNAT upload.

        :Note: This workflow is adapted from the AIRC workflow at
        https://everett.ohsu.edu/hg/qin_dce. The AIRC workflow time series
        merge is removed and added as input to the workflow created by this
        method. Any change to the ``qin_dce`` workflow should be reflected in
        this method.

        :param base_dir: the workflow working directory
        :param opts: the PK modeling parameters
        :return: the pyxnat Workflow
        """
        base_wf = pe.Workflow(name='modeling_base', base_dir=base_dir)

        # The PK modeling parameters.
        opts = self._pk_parameters(**opts)
        # Set the use_fixed_r1_0 flag.
        use_fixed_r1_0 = not not opts.get('r1_0_val')

        # Set up the input node.
        non_pk_flds = ['time_series', 'mask', 'bolus_arrival_index']
        in_fields = non_pk_flds + opts.keys()
        input_xfc = IdentityInterface(fields=in_fields, **opts)
        input_spec = pe.Node(input_xfc, name='input_spec')
        # Set the config parameters.
        for field in in_fields:
            if field in opts:
                setattr(input_spec.inputs, field, opts[field])

        # If we are not using a fixed r1_0 value, then compute a map from a
        # proton density weighted scan and the baseline of the DCE series.
        if not use_fixed_r1_0:
            # Create the DCE baseline image.
            make_base_func = Function(
                input_names=['time_series', 'baseline_end_idx'],
                output_names=['baseline'],
                function=make_baseline),
            make_base = pe.Node(make_base_func, name='make_base')
            base_wf.connect(input_spec, 'time_series',
                            make_base, 'time_series')
            base_wf.connect(input_spec, 'baseline_end_idx',
                            make_base, 'baseline_end_idx')

            # Create the R1_0 map.
            get_r1_0_func = Function(
                input_names=['pdw_image', 't1w_image', 'max_r1_0',
                             'mask_file'],
                output_names=['r1_0_map'],
                function=make_r1_0),
            get_r1_0 = pe.Node(get_r1_0_func, name='get_r1_0')
            base_wf.connect(input_spec, 'pd_nii', get_r1_0, 'pdw_image')
            base_wf.connect(make_base, 'baseline', get_r1_0, 't1w_image')
            base_wf.connect(input_spec, 'max_r1_0', get_r1_0, 'max_r1_0')
            base_wf.connect(input_spec, 'mask', get_r1_0, 'mask_file')

        # Convert the DCE time series to R1 maps.
        get_r1_series_func = Function(
            input_names=['time_series', 'r1_0', 'baseline_end', 'mask_file'],
            output_names=['r1_series'], function=make_r1_series)
        get_r1_series = pe.Node(get_r1_series_func, name='get_r1_series')
        base_wf.connect(input_spec, 'time_series',
                        get_r1_series, 'time_series')
        base_wf.connect(input_spec, 'baseline_end_idx',
                        get_r1_series, 'baseline_end')
        base_wf.connect(input_spec, 'mask', get_r1_series, 'mask_file')
        if use_fixed_r1_0:
            base_wf.connect(input_spec, 'r1_0_val', get_r1_series, 'r1_0')
        else:
            base_wf.connect(get_r1_0, 'r1_0_map', get_r1_series, 'r1_0')

        # Copy the time series meta-data to the R1 series.
        copy_meta = pe.Node(CopyMeta(), name='copy_meta')
        copy_meta.inputs.include_classes = [('global', 'const'),
                                            ('time', 'samples')]
        base_wf.connect(input_spec, 'time_series', copy_meta, 'src_file')
        base_wf.connect(get_r1_series, 'r1_series', copy_meta, 'dest_file')

        # Get the pharmacokinetic mapping parameters.
        get_params_flds = ['time_series', 'bolus_arrival_index']
        get_params_func = Function(input_names=get_params_flds,
                                   output_names=['params_csv'],
                                   function=get_fit_params)
        get_params = pe.Node(get_params_func, name='get_params')
        base_wf.connect(input_spec, 'time_series', get_params, 'time_series')
        base_wf.connect(input_spec, 'bolus_arrival_index',
                        get_params, 'bolus_arrival_index')

        # Import Fastfit on demand. This allows the modeling module to be
        # imported by applications without the proprietary Fastfit modeling
        # tool, e.g. the documentation builder Sphinx.
        try:
            Fastfit
        except NameError:
            from ..interfaces.fastfit import Fastfit
        # Optimize the pharmacokinetic model.
        pk_map = pe.Node(Fastfit(), name='pk_map')
        pk_map.inputs.model_name = 'fxr.model'
        pk_map.inputs.optional_outs = ['chisq', 'guess_model.k_trans',
                                       'guess_model.v_e', 'guess_model.chisq']
        base_wf.connect(copy_meta, 'dest_file', pk_map, 'target_data')
        base_wf.connect(input_spec, 'mask', pk_map, 'mask')
        base_wf.connect(get_params, 'params_csv', pk_map, 'params_csv')
        # Set the MPI flag.
        pk_map.inputs.use_mpi = DISTRIBUTABLE

        # Compute the K_trans difference.
        delta_k_trans = pe.Node(fsl.ImageMaths(), name='delta_k_trans')
        delta_k_trans.inputs.op_string = '-sub'
        base_wf.connect(pk_map, 'k_trans', delta_k_trans, 'in_file')
        base_wf.connect(pk_map, 'guess_model.k_trans',
                        delta_k_trans, 'in_file2')

        # Set up the outputs.
        outputs = ['r1_series', 'pk_params', 'fxr_k_trans', 'fxr_v_e',
                   'fxr_tau_i', 'fxr_chisq', 'fxl_k_trans', 'fxl_v_e',
                   'fxl_chisq', 'delta_k_trans']
        if not use_fixed_r1_0:
            outputs += ['dce_baseline', 'r1_0']
        output_spec = pe.Node(IdentityInterface(fields=outputs),
                              name='output_spec')
        base_wf.connect(copy_meta, 'dest_file', output_spec, 'r1_series')
        base_wf.connect(get_params, 'params_csv', output_spec, 'pk_params')
        base_wf.connect(pk_map, 'k_trans', output_spec, 'fxr_k_trans')
        base_wf.connect(pk_map, 'v_e', output_spec, 'fxr_v_e')
        base_wf.connect(pk_map, 'tau_i', output_spec, 'fxr_tau_i')
        base_wf.connect(pk_map, 'chisq', output_spec, 'fxr_chisq')
        base_wf.connect(pk_map, 'guess_model.k_trans',
                        output_spec, 'fxl_k_trans')
        base_wf.connect(pk_map, 'guess_model.v_e', output_spec, 'fxl_v_e')
        base_wf.connect(pk_map, 'guess_model.chisq', output_spec, 'fxl_chisq')
        base_wf.connect(delta_k_trans, 'out_file',
                        output_spec, 'delta_k_trans')

        if not use_fixed_r1_0:
            base_wf.connect(make_base, 'baseline',
                            output_spec, 'dce_baseline')
            base_wf.connect(get_r1_0, 'r1_0_map', output_spec, 'r1_0')

        self._configure_nodes(base_wf)

        return base_wf

    def _pk_parameters(self, **opts):
        """
        Collects the modeling parameters defined in either the options
        or the configuration, as described in :class:`ModelingWorkflow`.

        :param opts: the input options
        :return: the parameter {name: value} dictionary
        """
        config = self.configuration.get('Parameters', {})

        logger(__name__).debug("Setting the PK parameters from the option"
            " keyword parameters %s and configuration %s..." % (opts, config))

        # The R1_0 computation fields.
        r1_fields = ['pd_dir', 'max_r1_0']
        # All of the possible fields.
        fields = set(r1_fields).update(['baseline_end_idx', 'r1_0_val'])
        # Validate the input options.
        for field in opts.iterkeys():
            if field not in fields:
                raise KeyError("The PK paramter is not recognized: %s" % field)

        if 'baseline_end_idx' not in opts:
            # Look for the the baseline parameter in the configuration.
            if 'baseline_end_idx' in config:
                opts['baseline_end_idx'] = config['baseline_end_idx']
            else:
                # The default baseline image count.
                opts['baseline_end_idx'] = 1

        # Set the use_fixed_r1_0 variable to None, signifying unknown.
        use_fixed_r1_0 = None
        # Get the R1_0 parameter values.
        if 'r1_0_val' in opts:
            r1_0_val = opts.get('r1_0_val')
            if r1_0_val:
                use_fixed_r1_0 = True
            else:
                use_fixed_r1_0 = False
        else:
            for field in r1_fields:
                value = opts.get(field)
                if value:
                    use_fixed_r1_0 = False

        # If none of the R1_0 options are set in the options,
        # then try the configuration.
        if use_fixed_r1_0 == None:
            r1_0_val = config.get('r1_0_val')
            if r1_0_val:
                opts['r1_0_val'] = r1_0_val
                use_fixed_r1_0 = True

        # If R1_0 is not fixed, then augment the R1_0 options
        # from the configuration, if necessary.
        if not use_fixed_r1_0:
            for field in r1_fields:
                if field not in opts and field in config:
                    use_fixed_r1_0 = False
                    opts[field] = config[field]
                # Validate the R1 parameter.
                if not opts.get(field):
                    raise ValueError("Missing both the r1_0_val and the %s"
                        " parameter." % field)

        # If the use_fixed_r1_0 flag is set, then remove the
        # extraneous R1 computation fields.
        if use_fixed_r1_0:
            for field in r1_fields:
                opts.pop(field, None)

        self._logger.debug("The PK modeling parameters: %s" % opts)
        return opts


### Utility functions called by workflow nodes. ###

def make_baseline(time_series, baseline_end_idx):
    """
    Makes the R1_0 computation baseline NiFTI file.

    :param time_series: the modeling input 4D NiFTI image
    :param baseline_end_idx: the exclusive limit of the baseline
        computation input series
    :return: the baseline NiFTI file name
    :raise ValueError: if the end index is a negative number
    """
    from dcmstack.dcmmeta import NiftiWrapper

    if baseline_end_idx <= 0:
        raise ValueError("The R1_0 computation baseline end index input value"
                         " is not a positive number: %s" % baseline_end_idx)
    nii = nb.load(time_series)
    nw = NiftiWrapper(nii)

    baselines = []
    for idx, split_nii in nw.split():
        if idx == baseline_end_idx:
            break
        baselines.append(split_nii)

    if len(baselines) == 1:
        return baselines[0]
    else:
        baseline = NiftiWrapper.from_sequence(baselines)
    baseline_path = path.join(os.getcwd(), 'baseline.nii.gz')
    nb.save(baseline, baseline_path)

    return baseline_path


def make_r1_0(pdw_image, t1w_image, max_r1_0, mask_file, **kwargs):
    """
    Returns the R1_0 map NiFTI file from the given proton density
    and T1-weighted images. The R1_0 map is computed using the
    ``pdw_t1w_to_r1`` function. The ``pdw_t1w_to_r1`` module
    must be in the Python path.

    :param pdw_image: the proton density NiFTI image file path
    :param t1w_image: the T1-weighted image file path
    :param max_r1_0: the R1_0 range maximum
    :param mask_file: the modeling mask file path
    :param kwargs: the ``pdw_t1w_to_r1`` function keyword arguments
    :return: the R1_0 map NiFTI image file name
    """
    import os
    import nibabel as nb
    import numpy as np
    from pdw_t1w_to_r1 import pdw_t1w_to_r1
    from dcmstack.dcmmeta import NiftiWrapper

    pdw_nw = NiftiWrapper(nb.load(pdw_image), make_empty=True)
    t1w_nw = NiftiWrapper(nb.load(t1w_image), make_empty=True)
    r1_space = np.arange(0.01, max_r1_0, 0.01)
    mask = nb.load(mask_file).get_data()
    r1_0 = pdw_t1w_to_r1(pdw_nw, t1w_nw, r1_space=r1_space, mask=mask,
                         **kwargs)

    cwd = os.getcwd()
    out_nii = nb.Nifti1Image(r1_0, pdw_nw.nii_img.get_affine())
    out_fn = os.path.join(cwd, 'r1_0_map.nii.gz')
    nb.save(out_nii, out_fn)

    return out_fn


def make_r1_series(time_series, r1_0, mask_file, **kwargs):
    """
    Makes the R1_0 series NiFTI file.

    :param time_series: the modeling input 4D NiFTI image
    :param r1_0: the R1_0 image file path
    :param mask_file: the modeling mask file path
    :return: the R1_0 series NiFTI image file name
    """
    import os
    import nibabel as nb
    from dce_to_r1 import dce_to_r1
    from dcmstack.dcmmeta import NiftiWrapper

    dce_nw = NiftiWrapper(nb.load(time_series), make_empty=True)
    if not isinstance(r1_0, float):
        r1_0 = nb.load(r1_0).get_data()
    mask = nb.load(mask_file).get_data()
    r1_series = dce_to_r1(dce_nw, r1_0, mask=mask, **kwargs)

    cwd = os.getcwd()
    out_nii = nb.Nifti1Image(r1_series, dce_nw.nii_img.get_affine())
    out_fn = os.path.join(cwd, 'r1_series.nii.gz')
    nb.save(out_nii, out_fn)

    return out_fn


def get_fit_params(time_series, bolus_arrival_index):
    """
    Obtains the following modeling fit parameters:

    * *aif_params*: arterial input function parameter array
    * *aif_delta_t*: acquisition time deltas
    * *aif_shift*: acquisition time shift
    * *r1_cr*: contrast R1
    * *r1_b_pre*: pre-contrast R1

    :param time_series: the modeling input 4D NiFTI image
    :param bolus_arrival_index: the bolus uptake series index
    :return: the parameter CSV file path
    """
    import os
    import csv
    import nibabel as nb
    import numpy as np
    from dcmstack.dcmmeta import NiftiWrapper
    from dcmstack import dcm_time_to_sec

    # Load the time series into a nifty NiFTI wrapper.
    nii = nb.load(time_series)
    nw = NiftiWrapper(nii)

    # The AIF shift parameter is the offset of the bolus arrival
    # series mid-point acquisition time from the MR session start
    # time.
    acq_time0 = dcm_time_to_sec(nw.get_meta('AcquisitionTime', (0, 0, 0, 0)))
    acq_time1 = dcm_time_to_sec(
        nw.get_meta('AcquisitionTime', (0, 0, 0, bolus_arrival_index)))
    acq_time2 = dcm_time_to_sec(
        nw.get_meta('AcquisitionTime', (0, 0, 0, bolus_arrival_index + 1)))
    aif_shift = ((acq_time1 + acq_time2) / 2.0) - acq_time0

    # Create the parameter CSV file.
    with open('params.csv', 'w') as csv_file:
        csv_writer = csv.writer(csv_file)
        csv_writer.writerow(['aif_params', '0.674', '0.4', '2.2',
                             '0.23', '1.3', '0.09', '0.0013', '0.0'])
        csv_writer.writerow(['aif_delta_t', '1.5'])
        csv_writer.writerow(['aif_shift', str(aif_shift)])
        csv_writer.writerow(['r1_cr', '3.8'])
        csv_writer.writerow(['r1_b_pre', '0.71'])

    return os.path.join(os.getcwd(), 'params.csv')



from base64 import b64encode as encode


def generate_unique_name(name):
    """
    Makes a name that is unique to the given test name.
    
    :param name: the test name
    :return: the test subject name
    """
    return encode(name).strip('=')



from nipype.interfaces.base import (
    BaseInterface, BaseInterfaceInputSpec, traits,
    File, TraitedSpec)
from qipipe.staging.staging_error import StagingError
from qipipe.staging.fix_dicom import fix_dicom_headers


class FixDicomInputSpec(BaseInterfaceInputSpec):
    collection = traits.Str(desc='The image collection', mandatory=True)

    subject = traits.Str(desc='The subject name', mandatory=True)

    in_file = File(exists=True, desc='The input DICOM file', mandatory=True)


class FixDicomOutputSpec(TraitedSpec):
    out_file = File(desc="The modified output file", exists=True)


class FixDicom(BaseInterface):

    """The FixDicom interface wraps the
    :meth:`qipipe.staging.fix_dicom.fix_dicom_headers` function."""

    input_spec = FixDicomInputSpec

    output_spec = FixDicomOutputSpec

    def _run_interface(self, runtime):
        fixed = fix_dicom_headers(self.inputs.collection, self.inputs.subject,
                                  self.inputs.in_file)
        if len(fixed) != 1:
            raise StagingError("Fixed DICOM file count is not one: %s" % fixed)
        self._out_file = fixed[0]

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._out_file

        return outputs



"""
Maps the OHSU DICOM Patient IDs to the CTP Patient IDs.
"""

import os
from nipype.interfaces.base import (traits, BaseInterfaceInputSpec,
                                    TraitedSpec, BaseInterface, File,
                                    Directory)
from ..staging.map_ctp import map_ctp


class MapCTPInputSpec(BaseInterfaceInputSpec):
    collection = traits.Str(mandatory=True, desc='The collection name')

    subjects = traits.CList(traits.Str(), mandatory=True,
                               desc='The DICOM Patient IDs to map')

    dest = Directory(desc='The optional directory to write the map file '
                     '(default current directory)')


class MapCTPOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='The output properties file')


class MapCTP(BaseInterface):
    """
    The MapCTP interface wraps the
    :meth:`qipipe.interfaces.map_ctp.map_ctp` method.
    """

    input_spec = MapCTPInputSpec

    output_spec = MapCTPOutputSpec

    def _run_interface(self, runtime):
        self._out_file = map_ctp(self.inputs.collection, *self.inputs.subjects,
                                 dest=self.inputs.dest)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self._out_file
        return outputs



import os
import re
import tempfile
import networkx as nx
from qixnat import project
import qixnat
from qiutil.collections import EMPTY_DICT
from qiutil.ast_config import read_config
from .distributable import DISTRIBUTABLE


class WorkflowBase(object):

    """
    The WorkflowBase class is the base class for the QIN workflow
    wrapper classes.

    If the :mod:`qipipe.pipeline.distributable' ``DISTRIBUTABLE`` flag
    is set, then the execution is distributed using the
    `AIRC Grid Engine`_.

    The workflow plug-in arguments and node inputs can be specified in a
    :class:`qiutil.ast_config.ASTConfig` file. The standard
    configuration file name is the lower-case name of the ``WorkflowBase``
    subclass with ``.cfg`` extension, e.g. ``registration.cfg``. The
    configuration file paths to load in low-to-high precedence order
    consist of the following:

    1. the default config file ``default.cfg`` in the qipipe project
       ``conf`` directory

    2. the standard config file name in the qipipe project ``conf``
       directory

    3. the standard config file name in the current working directory

    4. the standard config file name in the ``QIN_CONF`` environment
       variable directory

    5. the *cfg_file* initialization parameter

    .. _AIRC Grid Engine: https://everett.ohsu.edu/wiki/GridEngine
    """

    CLASS_NAME_PAT = re.compile("^(\w+)Workflow$")
    """The workflow wrapper class name matcher."""

    DEF_CONF_DIR = os.path.join(os.path.dirname(__file__), '..', '..', 'conf')
    """The default configuration directory."""

    CFG_ENV_VAR='QIN_CONF'
    """The configuration directory environment variable."""

    INTERFACE_PREFIX_PAT = re.compile('(\w+\.)+interfaces?\.?')
    """
    Regexp matcher for an interface module.

    Example:

    >>> from qipipe.pipeline.workflow_base import WorkflowBase
    >>> WorkflowBase.INTERFACE_PREFIX_PAT.match('nipype.interfaces.ants.util.AverageImages').groups()
    ('nipype.',)
    """

    MODULE_PREFIX_PAT = re.compile('^((\w+\.)*)(\w+\.)(\w+)$')
    """
    Regexp matcher for a module prefix.

    Example:

    >>> from qipipe.pipeline.workflow_base import WorkflowBase
    >>> WorkflowBase.MODULE_PREFIX_PAT.match('ants.util.AverageImages').groups()
    ('ants.', 'ants.', 'util.', 'AverageImages')
    >>> WorkflowBase.MODULE_PREFIX_PAT.match('AverageImages')
    None
    """

    def __init__(self, logger, cfg_file=None, dry_run=False):
        """
        Initializes this workflow wrapper object.
        If the optional configuration file is specified, then the workflow
        settings in that file override the default settings.

        :param logger: the logger to use
        :param cfg_file: the optional workflow inputs configuration file
        :keyword dry_run: flag indicating whether to prepare but not
            run the pipeline
        """
        self._logger = logger

        self.configuration = self._load_configuration(cfg_file)
        """The workflow configuration."""

        self.dry_run = dry_run
        """Flag indicating whether to skip workflow job submission."""

    def depict_workflow(self, workflow):
        """
        Diagrams the given workflow graph. The diagram is written to the
        *name*\ ``.dot.png`` in the workflow base directory.

        :param workflow the workflow to diagram
        """
        fname = workflow.name + '.dot'
        if workflow.base_dir:
            grf = os.path.join(workflow.base_dir, fname)
        else:
            grf = fname
        workflow.write_graph(dotfilename=grf)
        self._logger.debug("The %s workflow graph is depicted at %s.png." %
                         (workflow.name, grf))

    def _load_configuration(self, cfg_file=None):
        """
        Loads the workflow configuration, as defined in
        :class:`qipipe.pipeline.workflow_base.WorkflowBase`.

        :param cfg_file: the optional configuration file path
        :return: the configuration dictionary
        """
        # The configuration files to load.
        cfg_files = []

        # The default configuration file.
        def_cfg_file = os.path.join(WorkflowBase.DEF_CONF_DIR, 'default.cfg')
        if os.path.exists(def_cfg_file):
            cfg_files.append(os.path.abspath(def_cfg_file))

        # Validate that the workflow name ends in Workflow.
        match = WorkflowBase.CLASS_NAME_PAT.match(self.__class__.__name__)
        if not match:
            raise NameError("The workflow wrapper class does not match the"
                            " standard workflow class name pattern: %s" %
                            self.__class__.__name__)
        name = match.group(1)
        fname = "%s.cfg" % name.lower()
        
        # The workflow-specific configuration file is in the conf directory.
        base_cfg_file = os.path.join(WorkflowBase.DEF_CONF_DIR, fname)
        if os.path.exists(base_cfg_file):
            cfg_files.append(os.path.abspath(base_cfg_file))

        # The working directory config file.
        cwd_cfg_file = os.path.abspath(fname)
        if os.path.exists(cwd_cfg_file) and cwd_cfg_file not in cfg_files:
            cfg_files.append(cwd_cfg_file)

        # The config file specified by the directory environment variable.
        env_cfg_dir = os.getenv(WorkflowBase.CFG_ENV_VAR, None)
        if env_cfg_dir:
            env_cfg_file = os.path.abspath(os.path.join(env_cfg_dir, fname))
            if os.path.exists(env_cfg_file) and env_cfg_file not in cfg_files:
                cfg_files.append(env_cfg_file)

        # The argument config file.
        if cfg_file:
            cfg_file = os.path.abspath(cfg_file)
            if os.path.exists(cfg_file) and cfg_file not in cfg_files:
                cfg_files.append(cfg_file)

        # Load the configuration.
        if cfg_files:
            self._logger.debug("Loading the %s configuration files %s..." %
                             (name, cfg_files))
            cfg = read_config(*cfg_files)
            return dict(cfg)
        else:
            return {}

    def _download_scans(self, xnat, subject, session, dest):
        """
        Download the NIFTI scan files for the given session.

        :param xnat: the :class:`qiutil.qixnat.XNAT` connection
        :param subject: the XNAT subject label
        :param session: the XNAT session label
        :param dest: the destination directory path
        :return: the download file paths
        """
        return xnat.download(project(), subject, session, dest=dest)

    def _run_workflow(self, workflow):
        """
        Executes the given workflow.

        :param workflow: the workflow to run
        """
        # If the workflow can be distributed, then get the plugin
        # arguments.
        if DISTRIBUTABLE:
            opts = self._configure_plugin(workflow)
        else:
            opts = {}

        # Set the base directory to an absolute path.
        if workflow.base_dir:
            workflow.base_dir = os.path.abspath(workflow.base_dir)
        else:
            workflow.base_dir = tempfile.mkdtemp()

        # Run the workflow.
        self._logger.debug("Executing the %s workflow in %s..." %
                           (workflow.name, workflow.base_dir))
        if self.dry_run:
            self._logger.debug("Skipped workflow %s job submission,"
                               " since the dry run flag is set." %
                               workflow.name)
        else:
            with qixnat.connect():
                workflow.run(**opts)

    def _inspect_workflow_inputs(self, workflow):
        """
        Collects the given workflow nodes' inputs for debugging.

        :return: a {node name: parameters} dictionary, where *parameters*
            is a node parameter {name: value} dictionary
        """
        return {node_name: self._inspect_node_inputs(workflow.get_node(node_name))
                for node_name in workflow.list_node_names()}

    def _inspect_node_inputs(self, node):
        """
        Collects the given node inputs and plugin arguments for debugging.

        :return: the node parameter {name: value} dictionary
        """
        fields = node.inputs.copyable_trait_names()
        param_dict = {}
        for field in fields:
            value = getattr(node.inputs, field)
            if value != None:
                param_dict[field] = value

        return param_dict

    def _configure_plugin(self, workflow):
        """
        Sets the *execution* and *SGE* parameters for the given workflow.
        See the ``conf`` directory files for examples.

        :param workflow: the workflow to run
        :return: the workflow execution arguments
        """
        # The workflow submission arguments:

        # The execution setting.
        if 'Execution' in self.configuration:
            workflow.config['execution'] = self.configuration['Execution']
            self._logger.debug("Workflow %s execution parameter: %s." %
                             (workflow.name, workflow.config['execution']))

        # The Grid Engine setting.
        if 'SGE' in self.configuration:
            opts = dict(plugin='SGE', **self.configuration['SGE'])
            self._logger.debug("Workflow %s plug-in parameters: %s." %
                             (workflow.name, opts))
        else:
            opts = {}

        return opts

    def _configure_nodes(self, workflow):
        """
        Sets the input parameters defined for the given workflow in
        this WorkflowBase's configuration. This method is called by
        each WorkflowBase subclass after the workflow is built and
        prior to execution.

        :Note: nested workflow nodes are not configured, e.g. if the
        ``registration`` workflow connects a `realign`` workflow
        node ``fnirt``, then the nested ``realign.fnirt`` node in
        ``registration`` is not configured.

        :param workflow: the workflow containing the nodes
        """
        # The default Grid Engine setting.
        if DISTRIBUTABLE and 'SGE' in self.configuration:
            def_plugin_args = self.configuration['SGE'].get('plugin_args')
            if def_plugin_args and 'qsub_args' in def_plugin_args:
                # Retain this workflow's default even if a node defined
                # in this workflow is included in a parent workflow.
                def_plugin_args['overwrite'] = True
                self._logger.debug("Workflow %s default node plug-in parameters:"
                                  " %s." % (workflow.name, def_plugin_args))
        else:
            def_plugin_args = None

        # The nodes defined in this workflow start with the workflow name.
        prefix = workflow.name + '.'
        # Configure all node inputs.
        nodes = [workflow.get_node(name)
                 for name in workflow.list_node_names()]
        for node in nodes:
            # An input {field: value} dictionary to format a debug log message.
            input_dict = {}
            # The node configuration.
            cfg = self._node_configuration(workflow, node)

            # Set the node inputs or plug-in argument.
            for attr, value in cfg.iteritems():
                if attr == 'plugin_args':
                    # If the workflow is on a cluster and the node plug-in
                    # arguments do not overwrite the default plug-in arguments,
                    # then append the node plug-in arguments to the default
                    # and set the overwrite flag. This ensures that the node
                    # plug-in arguments take precedence over the defaults and
                    # that the arguments are retained if the node is included
                    # in a parent workflow.
                    if DISTRIBUTABLE:
                        if ('qsub_args' in value and
                                'overwrite' not in value and
                                'qsub_args' in def_plugin_args):
                            qsub_args = value['qsub_args']
                            def_qsub_args = def_plugin_args['qsub_args']
                            value['qsub_args'] = def_qsub_args + ' ' + qsub_args
                            value['overwrite'] = True
                        setattr(node, attr, value)
                        self._logger.debug("%s workflow node %s plugin"
                                          " arguments: %s" %
                                          (workflow.name, node, value))
                elif value != getattr(node.inputs, attr):
                    # The config value differs from the default value.
                    # Capture the config entry for update.
                    input_dict[attr] = value
            
            # If:
            # 1) the configuration specifies a default,
            # 2) the node itself is not configured with plug-in arguments, and
            # 3) the node is defined in this workflow as opposed to a child
            #    workflow (i.e., the node name prefix is this workflow name),
            # then set the node plug-in arguments to the default.
            if (def_plugin_args and 'plugin_args' not in cfg and
                    str(node).startswith(prefix)):
                node.plugin_args = def_plugin_args

            # If a field was set to a config value, then print the config
            # setting to the log.
            if input_dict:
                self._set_inputs(node, input_dict)
                self._logger.debug("The following %s workflow node %s inputs"
                                   " were set from the configuration: %s" %
                                   (workflow.name, node, input_dict))

    def _set_inputs(self, node, input_dict):
        # Sort the input config fields by the requires dependency.
        traits = node.inputs.traits()
        attrs = set(input_dict)
        req_dict = {attr: set(traits[attr].requires).intersection(attrs)
                    for attr in input_dict
                    if traits[attr].requires}
        req_grf = nx.DiGraph()
        req_grf.add_nodes_from(input_dict)
        for attr, reqs in req_dict.iteritems():
            for req in reqs:
                req_grf.add_edge(req, attr)
        sorted_attrs = nx.topological_sort(req_grf)
        # Set the input config fields.
        for attr in sorted_attrs:
            setattr(node.inputs, attr, input_dict[attr])

    def _node_configuration(self, workflow, node):
        """
        Returns the {parameter: value} dictionary defined for the given
        node in this WorkflowBase's configuration. The configuration topic
        is determined as follows:

        * the node class, as described in :meth:`_interface_configuration`

        * the node name, qualified by the node hierarchy if the node is
          defined in a child workflow

        :param node: the interface class to check
        :return: the corresponding {field: value} dictionary
        """
        return (self._interface_configuration(node.interface.__class__) or
         self._node_name_configuration(workflow, node) or EMPTY_DICT)

    def _node_name_configuration(self, workflow, node):
        """
        Returns the {parameter: value} dictionary defined for the given
        node name, qualified by the node hierarchy if the node is
        defined in a child workflow.

        :param workflow: the active workflow
        :param node: the interface class to check
        :return: the corresponding {field: value} dictionary
        """
        if node._hierarchy == workflow.name:
            return self.configuration.get(node.name)
        else:
            return self.configuration.get(node.fullname)

    def _interface_configuration(self, klass):
        """
        Returns the {parameter: value} dictionary defined for the given
        interface class in this WorkflowBase's configuration. The
        configuration topic matches the module path of the interface class
        name. The topic can elide the ``interfaces`` or ``interface`` prefix,
        e.g.:

            [nipype.interfaces.ants.AverageImages]
            [ants.AverageImages]

        both refer to the ANTS AverageImages interface.

        :param node: the interface class to check
        :return: the corresponding {field: value} dictionary
        """
        topic = "%s.%s" % (klass.__module__,
                           klass.__name__)
        if topic in self.configuration:
            return self.configuration[topic]
        elif WorkflowBase.INTERFACE_PREFIX_PAT.match(topic):
            # A parent module might import the class. Therefore,
            # strip out the last module and retry.
            abbr = WorkflowBase.INTERFACE_PREFIX_PAT.sub('', topic)
            while abbr:
                if abbr in self.configuration:
                    return self.configuration[abbr]
                match = WorkflowBase.MODULE_PREFIX_PAT.match(abbr)
                if match:
                    prefix, _, _, name = match.groups()
                    abbr = prefix + name
                else:
                    abbr = None



import os
import logging
from collections import defaultdict
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import IdentityInterface
from nipype.interfaces.dcmstack import DcmStack
from qixnat import project
from ..interfaces import (Gate, FixDicom, Compress, XNATFind, XNATCopy)
import qixnat
from .workflow_base import WorkflowBase
from qiutil.logging import logger
from ..staging import staging_helper


def set_workflow_inputs(exec_wf, collection, subject, session,
                        scan_dict, dest=None):
    """
    Sets the given execution workflow inputs.
    The execution workflow must have the same input and iterable
    node names and fields as the :class:`StagingWorkflow` workflow.

    :param exec_wf: the workflow to execute
    :param collection: the AIRC collection name
    :param subject: the subject name
    :param session: the session name
    :param scan_dict: the *{scan number: [dicom files]}* dictionary
    :param dest: the TCIA staging destination directory (default is
        the current working directory)
    """
    # The input scans to stage.
    scans = scan_dict.keys()
    # Make a staging area subdirectory for each scan.
    stg_dict = _create_staging_area(subject, session, scans, dest)
    # The staging destination directories are pair-wise synchronized
    # with the input scans.
    dests = [stg_dict[scan] for scan in scans]
    iterables = dict(series=scans, dest=dests).items()

    # Set the top-level inputs.
    input_spec = exec_wf.get_node('input_spec')
    input_spec.inputs.collection = collection
    input_spec.inputs.subject = subject
    input_spec.inputs.session = session

    # Set the series iterator inputs.
    iter_series = exec_wf.get_node('iter_series')
    iter_series.iterables = iterables
    # Iterate over the series and dest input fields in lock-step.
    iter_series.synchronize = True

    # Set the DICOM file iterator inputs.
    iter_dicom = exec_wf.get_node('iter_dicom')
    iter_dicom.itersource = ('iter_series', 'series')
    iter_dicom.iterables = ('dicom_file', scan_dict)


def _create_staging_area(subject, session, scans, dest=None):
    """
    :param subject: the subject name
    :param session: the session name
    :param scans: the input scans
    :param dest: the TCIA staging destination directory (default is
        the current working directory)
    :return: the {scan number: directory} dictionary
    """
    # The staging location.
    dest = os.path.abspath(dest) if dest else os.getcwd()
    # Collect the {scan: destination} dictionary.
    return {scan: _make_series_staging_directory(dest, subject, session, scan)
            for scan in scans}


def _make_series_staging_directory(dest, subject, session, series):
    """
    Returns the dest/subject/session/series directory path in which to
    place DICOM files for TCIA upload. Creates the directory, if
    necessary.

    :return: the target series directory path
    """
    path = os.path.join(dest, subject, session, str(series))
    if not os.path.exists(path):
        os.makedirs(path)

    return path


class StagingWorkflow(WorkflowBase):

    """
    The StagingWorkflow class builds and executes the staging Nipype workflow.
    The staging workflow includes the following steps:

    - Group the input DICOM images into series.

    - Fix each input DICOM file header using the
      :class:`qipipe.interfaces.fix_dicom.FixDicom` interface.

    - Compress each corrected DICOM file.

    - Upload each compressed DICOM file into XNAT.

    - Stack each new series's 2-D DICOM files into a 3-D series NiFTI file
      using the DcmStack_ interface.

    - Upload each new series stack into XNAT.

    - Make the CTP_ QIN-to-TCIA subject id map.

    - Collect the id map and the compressed DICOM images into a target
      directory in collection/subject/session/series format for TCIA
      upload.

    The staging workflow input is the *input_spec* node consisting of
    the following input fields:

    - *collection*: the collection name

    - *subject*: the subject name

    - *session*: the session name

    The staging workflow has two iterables:

    - the *iter_series* node with input fields *series* and *dest*

    - the *iter_dicom* node with input fields *series* and *dicom_file*

    These iterables must be set prior to workflow execution. The
    *iter_series* *dest* input is the destination directory for
    the *iter_series* *series*.

    The *iter_dicom* node *itersource* is the ``iter_series.series``
    field. The ``iter_dicom.dicom_file`` iterables is set to the
    {series: [DICOM files]} dictionary.

    The DICOM files to upload to TCIA are placed in the destination
    directory in the following hierarchy:

        ``/path/to/dest/``
          *subject*\ /
            *session*\ /
              ``Series``\ *series_number*\ /
                *file*\ ``.dcm.gz``
                ...

    where:

    * *subject* is the subject name, e.g. ``Breast011``

    * *session* is the session name, e.g. ``Session03``

    * *series_number* is the DICOM Series Number

    * *file* is the DICOM file base name

    The staging workflow output is the *output_spec* node consisting
    of the following output field:

    - *image*: the session series stack NiFTI image file

    .. _CTP: https://wiki.cancerimagingarchive.net/display/Public/Image+Submitter+Site+User%27s+Guide
    .. _DcmStack: http://nipy.sourceforge.net/nipype/interfaces/generated/nipype.interfaces.dcmstack.html
    """

    def __init__(self, scan_type, **opts):
        """
        If the optional configuration file is specified, then the workflow
        settings in that file override the default settings.

        :param scan_type: the scan type, e.g. ``t1``
        :parameter opts: the following keword options:
        :keyword project: the XNAT project (default ``QIN``)
        :keyword base_dir: the workflow execution directory
            (default a new temp directory)
        :keyword cfg_file: the optional workflow inputs configuration file
        """
        cfg_file = opts.pop('cfg_file', None)
        super(StagingWorkflow, self).__init__(logger(__name__), cfg_file)

        # Set the XNAT project.
        prj = opts.pop('project', None)
        if prj:
            project(prj)
            self._logger.debug("Set the XNAT project to %s." % prj)

        # Make the workflow.
        self.workflow = self._create_workflow(scan_type, **opts)
        """
        The staging workflow sequence described in
        :class:`qipipe.pipeline.staging.StagingWorkflow`.
        """

    def set_inputs(self, collection, subject, session, scan_dict,
                   dest=None):
        """
        Sets the staging workflow inputs.

        :param collection: the collection name
        :param subject: the subject name
        :param session: the session name
        :param scan_dict: the *{series: [dicom files]}* dictionary
        :param dest: the TCIA staging destination directory (default is
            the current working directory)
        """
        set_workflow_inputs(self.workflow, collection, subject, session,
                            scan_dict, dest)

    def run(self):
        """Executes the staging workflow."""
        self._run_workflow(self.workflow)

    def _create_workflow(self, scan_type, base_dir=None):
        """
        Makes the staging workflow described in
        :class:`qipipe.pipeline.staging.StagingWorkflow`.

        :param scan_type: the scan type, e.g. ``t1``
        :param base_dir: the workflow execution directory
            (default is a new temp directory)
        :return: the new workflow
        """
        self._logger.debug("Creating the %s DICOM processing workflow..." %
                           scan_type)

        # The Nipype workflow object.
        workflow = pe.Workflow(name='staging', base_dir=base_dir)

        # The workflow input.
        in_fields = ['collection', 'subject', 'session']
        input_spec = pe.Node(IdentityInterface(fields=in_fields),
                             name='input_spec')
        self._logger.debug("The %s workflow input node is %s with fields %s" %
                         (workflow.name, input_spec.name, in_fields))

        # Create the session, if necessary.
        find_session_xfc = XNATFind(project=project(), create=True)
        find_session = pe.Node(find_session_xfc, name='find_session')
        workflow.connect(input_spec, 'subject', find_session, 'subject')
        workflow.connect(input_spec, 'session', find_session, 'session')

        # The series iterator.
        iter_series_fields = ['series', 'dest']
        iter_series = pe.Node(IdentityInterface(fields=iter_series_fields),
                              name='iter_series')
        self._logger.debug("The %s workflow series iterable node is %s with"
                           " fields %s" % (workflow.name, iter_series.name,
                                           iter_series_fields))
        
        # The DICOM file iterator.
        iter_dicom_fields = ['series', 'dicom_file']
        iter_dicom = pe.Node(IdentityInterface(fields=iter_dicom_fields),
                             name='iter_dicom')
        self._logger.debug("The %s workflow DICOM iterable node is %s with"
                           " iterable source %s and iterables"
                           " ('%s', {%s: [%s]})" %
                           (workflow.name, iter_dicom.name, iter_series.name,
                           'dicom_file', 'series', 'DICOM files'))
        workflow.connect(iter_series, 'series', iter_dicom, 'series')

        # Fix the AIRC DICOM tags.
        fix_dicom = pe.Node(FixDicom(), name='fix_dicom')
        workflow.connect(input_spec, 'collection', fix_dicom, 'collection')
        workflow.connect(input_spec, 'subject', fix_dicom, 'subject')
        workflow.connect(iter_dicom, 'dicom_file', fix_dicom, 'in_file')
        
        # If the scan type is T1, then compress the corrected DICOM files.
        # T2 scan DICOM files omit DICOM compress and upload.
        if scan_type == 't1':
            compress_dicom = pe.Node(Compress(), name='compress_dicom')
            workflow.connect(fix_dicom, 'out_file', compress_dicom, 'in_file')
            workflow.connect(iter_series, 'dest', compress_dicom, 'dest')

            # Force the DICOM upload to follow session create.
            # Since only one upload task can run at a time for a given series,
            # this upload gate node is a JoinNode that collects the iterated
            # scan DICOM files.
            upload_dicom_gate_xfc = Gate(fields=['xnat_id', 'scan', 'files'])
            upload_dicom_gate = pe.JoinNode(upload_dicom_gate_xfc,
                                            joinsource='iter_dicom',
                                            joinfield='files',
                                            name='upload_dicom_gate')
            workflow.connect(find_session, 'xnat_id', upload_dicom_gate, 'xnat_id')
            workflow.connect(iter_series, 'series', upload_dicom_gate, 'scan')
            workflow.connect(compress_dicom, 'out_file', upload_dicom_gate, 'files')

            upload_dicom_xfc = XNATCopy(project=project(), resource='DICOM',
                                          skip_existing=True)
            upload_dicom = pe.Node(upload_dicom_xfc, name='upload_dicom')
            workflow.connect(input_spec, 'subject', upload_dicom, 'subject')
            workflow.connect(input_spec, 'session', upload_dicom, 'session')
            workflow.connect(upload_dicom_gate, 'scan', upload_dicom, 'scan')
            workflow.connect(upload_dicom_gate, 'files', upload_dicom, 'in_files')

        # Stack the scan into a 3D NiFTI file.
        suffix = "_%s" % scan_type
        stack_xfc = DcmStack(embed_meta=True,
                             out_format="series%(SeriesNumber)03d" + suffix)
        stack = pe.JoinNode(stack_xfc, joinsource='iter_dicom',
                            joinfield='dicom_files', name='stack')
        
        workflow.connect(fix_dicom, 'out_file', stack, 'dicom_files')

        # Force the T1 3D upload to follow DICOM upload.
        # Note: XNAT fails app. 80% into the T1 upload. It appears to be a
        # concurrency conflict, possibly arising from the following causes:
        # * the non-reentrant pyxnat's custom non-http2lib cache is corrupted
        # * an XNAT archive directory access race condition
        #
        # However, the error cannot be isolated for the following reasons:
        # * the error is sporadic and unreproducible
        # * since nipype swallows non-nipype Python log messages, the upload
        #   and pyxnat log messages disappear
        #
        # This gate task serializes upload to prevent potential XNAT access
        # conflicts.
        #
        # TODO - isolate and fix.
        #
        if scan_type == 't1':
            upload_3d_gate_xfc = Gate(fields=['out_file', 'xnat_files'])
            upload_3d_gate = pe.Node(upload_3d_gate_xfc, name='upload_3d_gate')
            workflow.connect(upload_dicom, 'xnat_files', upload_3d_gate, 'xnat_files')
            workflow.connect(stack, 'out_file', upload_3d_gate, 'out_file')

        # Upload the 3D NiFTI stack files to XNAT.
        upload_3d_xfc = XNATCopy(project=project(), resource='NIFTI',
                                 skip_existing=True)
        upload_3d = pe.Node(upload_3d_xfc, name='upload_3d')
        workflow.connect(input_spec, 'subject', upload_3d, 'subject')
        workflow.connect(input_spec, 'session', upload_3d, 'session')
        workflow.connect(iter_series, 'series', upload_3d, 'scan')
        # T1 upload is gated by DICOM upload.
        if scan_type == 't1':
            workflow.connect(upload_3d_gate, 'out_file', upload_3d, 'in_files')
        else:
            workflow.connect(stack, 'out_file', upload_3d, 'in_files')

        # The output is the 3D NiFTI stack file.
        output_spec = pe.Node(Gate(fields=['image', 'xnat_files']), name='output_spec')
        workflow.connect(stack, 'out_file', output_spec, 'image')
        workflow.connect(upload_3d, 'xnat_files', output_spec, 'xnat_files')

        self._configure_nodes(workflow)

        self._logger.debug("Created the %s workflow." % workflow.name)
        # If debug is set, then diagram the workflow graph.
        if self._logger.level <= logging.DEBUG:
            self.depict_workflow(workflow)

        return workflow



import os
from nipype.interfaces.base import (
    traits, CommandLine, CommandLineInputSpec, TraitedSpec, InputMultiPath,
    File)
from nipype.interfaces.traits_extension import isdefined


class XNATCopyInputSpec(CommandLineInputSpec):
    """
    The input spec with arguments in the following order:
    * options
    * the input files, for an upload
    * the XNAT object path
    * the destination directory, for a download
    """

    project = traits.Str(argstr='--project %s',
                         desc='The XNAT project id')

    force = traits.Bool(argstr='--force',
                        desc='Flag indicating whether to replace an existing'
                             ' XNAT file')

    skip_existing = traits.Bool(argstr='--skip-existing',
                                desc='Flag indicating whether to skip upload'
                                     ' to an existing target XNAT file')

    # The input files to upload precede the XNAT path.
    # The input file arguments follow the options and precede the target
    # XNAT path.
    in_files = InputMultiPath(File(exists=True), argstr='%s', position=-3,
                              desc='The files to upload')

    subject = traits.Str(mandatory=True, desc='The XNAT subject name')

    # The XNAT path masquerades as the session.
    # The XNAT path argument follows the input files for an upload and precedes
    # the destination for a download.
    session = traits.Str(mandatory=True, argstr='%s', position=-2,
                         desc='The XNAT session name')

    resource = traits.Str(desc='The XNAT resource name (scan default is NIFTI)')

    scan = traits.Either(traits.Int, traits.Str, desc='The XNAT scan name')

    reconstruction = traits.Str(desc='The XNAT reconstruction name')

    assessor = traits.Str(desc='The XNAT assessor name')

    # The download destination follows the XNAT path.
    # The destination goes at the end of the command line.
    dest = traits.Str(argstr='%s', position=-1, desc='The download directory')


class XNATCopyOutputSpec(TraitedSpec):
    xnat_files = traits.List(traits.Str, desc='The XNAT file object labels')


class XNATCopy(CommandLine):
    """
    The ``XNATCopy`` Nipype interface wraps the ``qicp`` command.
    """

    input_spec = XNATCopyInputSpec

    output_spec = XNATCopyOutputSpec

    cmd = 'qicp'

    def _format_path(self):
        """:return: the XNAT path prefixed by ``xnat:``"""

        # The XNAT object hierarchy starts with the subject and session.
        path = [self.inputs.subject, self.inputs.session]

        # The resource parent container.
        if isdefined(self.inputs.scan):
            path.append('scan')
            path.append(str(self.inputs.scan))
        elif isdefined(self.inputs.reconstruction):
            path.append('reconstruction')
            path.append(self.inputs.reconstruction)
        elif isdefined(self.inputs.assessor):
            path.append('assessor')
            path.append(self.inputs.assessor)

        # The resource.
        if isdefined(self.inputs.resource):
            path.append('resource')
            path.append(self.inputs.resource)

        # The path is terminated with 'files'.
        path.append('files')

        # Make the path string prefixed by xnat:.
        return 'xnat:' + '/'.join(path)

    def _format_arg(self, opt, spec, val):
        if opt == 'session':
            return self._format_path()
        else:
            return super(XNATCopy, self)._format_arg(opt, spec, val)


    def _list_outputs(self):
        outputs = self._outputs().get()
        if isdefined(self.inputs.in_files):
            # The upload outputs the XNAT file names.
            outputs['xnat_files'] = [os.path.split(f)[1]
                                     for f in self.inputs.in_files]

        return outputs



import os
import shutil
import distutils
from nose.tools import (assert_equal, assert_is_not_none, assert_true)
from qipipe.pipeline import qipipeline as qip
import qixnat
from qipipe.staging import airc_collection as airc
from qipipe.staging.staging_helper import get_subjects
from qiutil.ast_config import read_config
from ... import (project, ROOT)
from ...helpers.logging import logger
from ...unit.pipeline.test_mask import MASK_CONF
from ...unit.pipeline.test_registration import REG_CONF
from ...unit.pipeline.test_modeling import MODELING_CONF

REG_CONF = os.path.join(ROOT, 'conf', 'registration.cfg')
"""The test registration configuration."""

RESULTS = os.path.join(ROOT, 'results', 'pipeline', 'qipipeline')
"""The test results directory."""

FIXTURES = os.path.join(ROOT, 'fixtures', 'staging')
"""The test fixture directory."""


class TestQIPipeline(object):

    """
    QIN Pipeline unit tests.
    
    Note:: a precondition for running this test is that the environment
        variable ``QIN_DATA`` is set to the AIRC ``HUANG_LAB`` mount point.
        If ``QIN_DATA`` is not set, then no test cases are run and a
        log message is issued.
    
    Note:: the modeling workflow is only executed if the ``fastfit``
        executable is found.
    
    Note:: this test takes app. four days to run serially without modeling.
    """

    def setUp(self):
        shutil.rmtree(RESULTS, True)

    def tearDown(self):
        shutil.rmtree(RESULTS, True)

    def test_breast_scans(self):
        data = os.getenv('QIN_DATA')
        if data:
            fixture = os.path.join(RESULTS, 'data', 'breast')
            parent = os.path.join(fixture, 'BreastChemo1')
            os.makedirs(parent)
            src = os.path.join(data, 'Breast_Chemo_Study', 'BreastChemo3',
                               'Visit1')
            assert_true(os.path.exists(src), "Breast test fixture not found:"
                        " %s" % src)
            dest = os.path.join(parent, 'Visit1')
            os.symlink(src, dest)
            self._test_collection('Breast', fixture)
        else:
            logger(__name__).info("Skipping the QIN pipeline unit Breast"
                                  " test, since the QIN_DATA environment"
                                  " variable is not set.")

    def test_sarcoma_scans(self):
        data = os.getenv('QIN_DATA')
        if data:
            fixture = os.path.join(RESULTS, 'data', 'sarcoma')
            parent = os.path.join(fixture, 'Subj_1')
            os.makedirs(parent)
            src = os.path.join(data, 'Sarcoma', 'Subj_1', 'Visit_1')
            assert_true(os.path.exists(src), "Sarcoma test fixture not found:"
                        " %s" % src)
            dest = os.path.join(parent, 'Visit_1')
            os.symlink(src, dest)
            self._test_collection('Sarcoma', fixture)
        else:
            logger(__name__).info("Skipping the QIN pipeline unit Sarcoma"
                                  " test, since the QIN_DATA environment"
                                  " variable is not set.")

    def _test_collection_scans(self, collection, fixture):
        """
        Run the pipeline on the given collection and verify that scans are
        created in XNAT.
        
        :param collection: the AIRC collection name
        :param fixture: the test input
        """
        logger(__name__).debug("Testing the QIN pipeline on %s..." % fixture)

        # The staging destination and work area.
        dest = os.path.join(RESULTS, 'data')
        base_dir = os.path.join(RESULTS, 'work')

        # The pipeline options.
        opts = dict(base_dir=base_dir, dest=dest, mask=MASK_CONF,
                    registration=REG_CONF)
        # If fastfit is not available, then only execute the staging and
        # registration workflows. Otherwise, execute all workflows.
        if not distutils.spawn.find_executable('fastfit'):
            opts['actions'] = ['stage', 'register']

        # The test subject => directory dictionary.
        sbj_dir_dict = get_subjects(collection, fixture)
        # The test subjects.
        subjects = sbj_dir_dict.keys()
        # The test source directories.
        sources = sbj_dir_dict.values()

        with qixnat.connect() as xnat:
            # Delete any existing test subjects.
            xnat.delete_subjects(project(), *subjects)

            # Run the staging, mask and registration workflows, but not
            # the modeling.
            logger(__name__).debug("Executing the QIN pipeline...")
            output_dict = qip.run(collection, *sources, **opts)
            # Verify the result.
            recon = qip
            for sbj, sess_dict in output_dict.iteritems():
                for sess, results in sess_dict.iteritems():
                    if opts['mask'] == False:
                        continue
                    # Verify the registration resource.
                    if opts['registration'] == False:
                        continue
                    # The XNAT registration resource name.
                    rsc = results['registration']
                    assert_is_not_none(rsc, 
                                       "The %s %s result does not have a"
                                       " registration resource" %
                                       (sbj, sess))
                    reg_obj = xnat.get_resource(
                        project(), sbj, sess, resource=rsc)
                    assert_true(reg_obj.exists(),
                                "The %s %s registration resource %s was not"
                                " created in XNAT" % (sbj, sess, rsc))
                    # Verify the modeling resource.
                    if opts['modeling'] != False:
                        rsc = results['modeling']
                        mdl_obj = xnat.get_resource(project(), sbj, sess, rsc)
                        assert_true(mdl_obj.exists(),
                                    "The %s %s modeling resource %s was not"
                                    " created in XNAT" % (sbj, sess, rsc))

            # Delete the test subjects.
            xnat.delete_subjects(project(), *subjects)


if __name__ == "__main__":
    import nose

    nose.main(defaultTest=__name__)



import os
import gzip
from nipype.interfaces.base import (BaseInterfaceInputSpec, TraitedSpec,
                                    BaseInterface, File, Directory)


class CompressInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc='The file to compress')

    dest = Directory(desc='The optional directory to write the compressed file'
                     '(default current directory)')


class CompressOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='The compressed file')


class Compress(BaseInterface):
    input_spec = CompressInputSpec

    output_spec = CompressOutputSpec

    def _run_interface(self, runtime):
        self.out_file = self._compress(
            self.inputs.in_file, dest=self.inputs.dest)

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self.out_file

        return outputs

    def _compress(self, in_file, dest=None):
        """
        Compresses the given file.
    
        :param in_file: the path of the file to compress
        :param dest: the destination (default is the working directory)
        :return: the compressed file path
        """
        if not dest:
            dest = os.getcwd()
        if not os.path.exists(dest):
            os.makedirs(dest)
        _, fname = os.path.split(in_file)
        out_file = os.path.join(dest, fname + '.gz')
        f = open(in_file, 'rb')
        cf = gzip.open(out_file, 'wb')
        cf.writelines(f)
        f.close()
        cf.close()

        return os.path.abspath(out_file)



from nipype.interfaces.base import (traits, BaseInterfaceInputSpec,
                                    TraitedSpec, BaseInterface)
from nipype.interfaces.traits_extension import isdefined
import qixnat


class XNATFindInputSpec(BaseInterfaceInputSpec):
    project = traits.Str(mandatory=True, desc='The XNAT project id')

    subject = traits.Str(mandatory=True, desc='The XNAT subject name')

    session = traits.Str(desc='The XNAT session name')

    scan = traits.Either(traits.Int, traits.Str, desc='The XNAT scan name')

    reconstruction = traits.Str(desc='The XNAT reconstruction name')

    assessor = traits.Str(desc='The XNAT assessor name')

    resource = traits.Str(desc='The XNAT resource name')

    file = traits.Str(desc='The XNAT file name')

    inout = traits.Str(desc='The XNAT resource in/out designator')

    create = traits.Bool(default=False, desc='Flag indicating whether to '
                         'create the XNAT object if it does not yet exist')


class XNATFindOutputSpec(TraitedSpec):
    xnat_id = traits.Str(desc='The XNAT object id')


class XNATFind(BaseInterface):

    """
    The ``XNATFind`` Nipype interface wraps the
    :meth:`qiutil.qixnat.find` method.
    
    :Note: only one XNAT operation can run at a time.
    """

    input_spec = XNATFindInputSpec

    output_spec = XNATFindOutputSpec

    def __init__(self, **inputs):
        super(XNATFind, self).__init__(**inputs)

    def _run_interface(self, runtime):
        # The find options.
        opts = dict(create=self.inputs.create)

        # The session is optional.
        if isdefined(self.inputs.session):
            session = self.inputs.session
        else:
            session = None

        # The resource parent.
        if self.inputs.scan:
            opts['modality'] = 'MR'
            opts['scan'] = self.inputs.scan
        elif self.inputs.reconstruction:
            opts['reconstruction'] = self.inputs.reconstruction
        elif self.inputs.assessor:
            opts['assessor'] = self.inputs.assessor

        # The resource.
        if isdefined(self.inputs.resource):
            opts['resource'] = self.inputs.resource

        # The file.
        if isdefined(self.inputs.file):
            opts['file'] = self.inputs.file

        # Delegate to the XNAT helper.
        with qixnat.connect() as xnat:
            obj = xnat.find(self.inputs.project, self.inputs.subject,
                            session, **opts)
            if obj and obj.exists():
                self._xnat_id = obj.id()
            else:
                self._xnat_id = None

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        if self._xnat_id:
            outputs['xnat_id'] = self._xnat_id

        return outputs



import re
from .staging_error import StagingError


def collection_with_name(name):
    """
    :param name: the OHSU QIN collection name
    :return: the corresponding AIRC collection
    :raise ValueError: if the given collection name is not recognized
    """
    if not hasattr(collection_with_name, 'extent'):
        setattr(collection_with_name, 'extent', _create_collections())
    if name not in collection_with_name.extent:
        raise ValueError(
            "The AIRC collection name is not recognized: %s" % name)

    return collection_with_name.extent[name]


def _create_collections():
    """Creates the pre-defined AIRC collections."""

    # The AIRC T1 scan DICOM files are in the concat directory.
    # The AIRC T2 scan DICOM files are in the sorted subdirectories. 
    dcm_pat_dict = dict(t1='*concat*/*', t2='*sorted*/*/*')

    return dict(
        Breast=AIRCCollection(
            'Breast', 'BreastChemo(\d+)', 'Visit(\d+)', dcm_pat_dict),
        Sarcoma=AIRCCollection(
            'Sarcoma', 'Subj_(\d+)', '(?:Visit_|S\d+V)(\d+)', dcm_pat_dict))


class AIRCCollection(object):

    """The AIRC Study characteristics."""

    def __init__(self, name, subject_pattern, session_pattern, dcm_pat_dict):
        """
        :param name: `self.name`
        :param subject_pattern: `self.subject_pattern`
        :param session_pattern: `self.session_pattern`
        :param dicom_pat_dict: `self.dicom_pat_dict`
        """
        self.name = name
        """The collection name."""

        self.subject_pattern = subject_pattern
        """The subject directory name match regular expression pattern."""

        self.session_pattern = session_pattern
        """The session directory name match regular expression pattern."""

        self.dcm_pat_dict = dcm_pat_dict
        """The {type: DICOM directory name match glob pattern} dictionary."""

    def path2subject_number(self, path):
        """
        :param path: the directory path
        :return: the subject number
        :raise StagingError: if the path does not match the collection subject
            pattern
        """
        match = re.search(self.subject_pattern, path)
        if not match:
            raise StagingError(
                "The directory path %s does not match the subject pattern %s" %
                (path, self.subject_pattern))

        return int(match.group(1))

    def path2session_number(self, path):
        """
        :param path: the directory path
        :return: the session number
        :raise StagingError: if the path does not match the collection session
            pattern
        """
        match = re.search(self.session_pattern, path)
        if not match:
            raise StagingError(
                "The directory path %s does not match the session pattern %s" %
                (path, self.session_pattern))
        return int(match.group(1))



import os
import re
import glob
import shutil
from nose.tools import (assert_equal, assert_true)
import nipype.pipeline.engine as pe
try:
    from qipipe.pipeline import modeling
except ImportError:
    modeling = None
from ... import (project, ROOT)
from ...helpers.logging import logger
from ...unit.pipeline.staged_test_base import StagedTestBase

MODELING_CONF = os.path.join(ROOT, 'conf', 'modeling.cfg')
"""The test registration configuration."""

FIXTURES = os.path.join(ROOT, 'fixtures', 'pipeline', 'modeling')

RESULTS = os.path.join(ROOT, 'results', 'pipeline', 'modeling')
"""The test results directory."""


class TestModelingWorkflow(StagedTestBase):

    """
    Modeling workflow unit tests.
    This test exercises the modeling workflow on the QIN Breast and Sarcoma
    study visits in the ``test/fixtures/pipeline/modeling`` test fixture
    directory.
    
    Note:: a precondition for running this test is that the
        ``test/fixtures/pipeline/modeling`` directory contains the series
        stack test data in collection/subject/session format, e.g.::
    
            breast
                Breast003
                    Session01
                        series009.nii.gz
                        series023.nii.gz
                         ...
            sarcoma
                Sarcoma001
                    Session01
                        series011.nii.gz
                        series013.nii.gz
                         ...
    
    The fixture is not included in the Git source repository due to storage
    constraints.
    
    Note:: this test takes app. 8 hours to run on the AIRC cluster.
    """

    def __init__(self):
        super(TestModelingWorkflow, self).__init__(
            logger(__name__), FIXTURES, RESULTS)

    def test_breast(self):
        if modeling:
            self._test_breast()
        else:
            logger(__name__).debug('Skipping modeling test since fastfit'
                                   ' is unavailable.')

    def test_sarcoma(self):
        if modeling:
            self._test_sarcoma()

    def _run_workflow(self, subject, session, *images, **opts):
        """
        Executes :meth:`qipipe.pipeline.modeling.run` on the input session scans.
        
        :param subject: the input subject
        :param session: the input session
        :param images: the input 3D NiFTI images to model
        :param opts: the  workflow options
        :return: the :meth:`qipipe.pipeline.modeling.run` result
        """
        # Run the workflow.
        return modeling.run(subject, session, *images, **opts)

    def _verify_result(self, xnat, subject, session, result):
        anl_obj = xnat.get_analysis(project(), sbj, sess, result)
        assert_true(anl_obj.exists(),
                    "The %s %s %s XNAT analysis object was not created" %
                    (sbj, sess, result))


if __name__ == "__main__":
    import nose

    nose.main(defaultTest=__name__)



import os
import gzip
from nipype.interfaces.base import (BaseInterfaceInputSpec, TraitedSpec,
                                    BaseInterface, File, Directory)


class UncompressInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True,
                   desc='The file to uncompress')

    dest = Directory(
        desc='The optional directory to write the uncompressed file (default current directory)')


class UncompressOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='The uncompressed file')


class Uncompress(BaseInterface):
    input_spec = UncompressInputSpec

    output_spec = UncompressOutputSpec

    def _run_interface(self, runtime):
        self.out_file = self._uncompress(
            self.inputs.in_file, dest=self.inputs.dest)
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self.out_file
        return outputs

    def _uncompress(self, in_file, dest=None):
        """
        Uncompresses the given file.
    
        :param in_file: the path of the file to uncompress
        @parma dest: the destination (default is the working directory)
        :return: the compressed file path
        """
        if not dest:
            dest = os.getcwd()
        if not os.path.exists(dest):
            os.makedirs(dest)
        _, fname = os.path.split(in_file)
        out_file = os.path.join(dest, fname[:-3])
        cf = gzip.open(in_file, 'rb')
        f = open(out_file, 'wb')
        f.writelines(cf)
        f.close()
        cf.close()
        return os.path.abspath(out_file)



"""Pipeline utility functions."""

import os
import re
import glob
from collections import defaultdict
from qiutil.collections import nested_defaultdict
from qiutil.logging import logger
import qixnat
from qixnat import project
from qidicom import reader
from qidicom.hierarchy import group_dicom_files_by_series
from . import airc_collection as airc
from .staging_error import StagingError

SUBJECT_FMT = '%s%03d'
"""The QIN subject name format with arguments (collection, subject number)."""

SESSION_FMT = 'Session%02d'
"""The QIN series name format with arguments (subject, series number)."""


def iter_stage(collection, *inputs, **opts):
    """
    Iterates over the the new AIRC visits in the given input directories.
    This method is a staging generator which yields a tuple consisting
    of the subject, session and a {scan type: {scan number: [DICOM files]}}
    dictionary, e.g.::
    
        >> sbj, sess, scan_type_dict = next(iter_stage('Sarcoma', '/path/to/subject'))
        >> print sbj
        Sarcoma001
        >> print sess
        Session01
        >> print scan_type_dict
        {'t1': {20: ['/path/to/file', ...]}, ...}

    The input directories conform to the
    :class:`qipipe.staging.airc_collection.AIRCCollection` *subject_pattern*)

    :param collection: the AIRC image collection name
    :param inputs: the AIRC source subject directories to stage
    :param opts: the following keyword option:
    :keyword dest: the TCIA staging destination directory (default is
        the current working directory)
    :keyword skip_existing: flag indicating whether to ignore existing
        sessions (default True)
    :yield: stage the DICOM files
    :yieldparam subject: the subject name
    :yieldparam session: the session name
    :yieldparam scan_type_dict: the *{scan type: {scan number: [DICOM files]}}*
        dictionary
    """
    # Validate that there is a collection
    if not collection:
        raise ValueError('Staging is missing the AIRC collection name')

    # The staging location.
    dest = opts.pop('dest', None)
    if dest:
        dest = os.path.abspath(dest)
    else:
        dest = os.getcwd()
    
    # Group the new DICOM files into a
    # {subject: {session: [(series, dicom_files), ...]}} dictionary.
    stg_dict = _collect_visits(collection, *inputs, **opts)
    if not stg_dict:
        return

    # The workflow subjects.
    subjects = stg_dict.keys()

    # Print a debug message.
    series_cnt = sum(map(len, stg_dict.itervalues()))
    logger(__name__).debug("Staging %d new %s series from %d subjects in"
                           " %s..." % (series_cnt, collection, len(subjects),
                                       dest))
    
    # Generate the (subject, session, {scan type: {series: [dicom files]}})
    # tuples.
    for sbj, sess_dict in stg_dict.iteritems():
        logger(__name__).debug("Staging subject %s..." % sbj)
        for sess, scan_type_dict in sess_dict.iteritems():
            logger(__name__).debug("Staging %s session %s..." % (sbj, sess))
            # Delegate to the workflow executor.
            yield sbj, sess, scan_type_dict
            logger(__name__).debug("Staged %s session %s." % (sbj, sess))
        logger(__name__).debug("Staged subject %s." % sbj)
    logger(__name__).debug("Staged %d new %s series from %d subjects in %s." %
                           (series_cnt, collection, len(subjects), dest))


def subject_for_directory(collection, path):
    """
    Infers the XNAT subject names from the given directory name.

    :param collection: the AIRC collection name
    :return: the corresponding XNAT subject label
    :raise StagingError: if the name does not match the collection subject
        pattern
    """
    airc_coll = airc.collection_with_name(collection)
    sbj_nbr = airc_coll.path2subject_number(path)
    return SUBJECT_FMT % (collection, sbj_nbr)


def get_subjects(collection, source, pattern=None):
    """
    Infers the XNAT subject names from the given source directory.
    The source directory contains subject subdirectories.
    The match pattern matches on the subdirectories and captures the
    subject number. The subject name is the collection name followed
    by the subject number, e.g. ``Breast004``.

    :param collection: the AIRC collection name
    :param source: the input parent directory
    :param pattern: the subject directory name match pattern (default
        is the :class:`qipipe.staging.airc_collection.AIRCCollection`
        *subject_pattern*)
    :return: the subject name => directory dictionary
    """
    logger(__name__).debug("Detecting %s subjects from %s..." %
          (collection, source))
    airc_coll = airc.collection_with_name(collection)
    pat = pattern or airc_coll.subject_pattern
    sbj_dir_dict = {}
    for d in os.listdir(source):
        match = re.match(pat, d)
        if match:
            # The XNAT subject name.
            subject = SUBJECT_FMT % (collection, int(match.group(1)))
            # The subject source directory.
            sbj_dir_dict[subject] = os.path.join(source, d)
            logger(__name__).debug(
                "Discovered XNAT test subject %s subdirectory: %s" %
                (subject, d))

    return sbj_dir_dict


def _collect_visits(collection, *inputs, **opts):
    """
    Collects the AIRC visits in the given input directories.
    The visit DICOM files are grouped by series.

    :param collection: the AIRC image collection name
    :param inputs: the AIRC source subject directories
    :param opts: the :class:`VisitIterator` initializer options,
        as well as the following keyword option:
    :keyword skip_existing: flag indicating whether to ignore existing
        sessions (default True)
    :return: the *{subject: {session: {scan type: [dicom files]}}}*
        dictionary
    """
    if opts.pop('skip_existing', True):
        visits = _detect_new_visits(collection, *inputs, **opts)
    else:
        visits = list(_iter_visits(collection, *inputs, **opts))

    # Group the DICOM files by series.
    return _group_visits(*visits)


def _detect_new_visits(collection, *inputs, **opts):
    """
    Detects the new AIRC visits in the given input directories. The visit
    DICOM files are grouped by series.

    :param collection: the AIRC image collection name
    :param inputs: the AIRC source subject directories
    :param opts: the :class:`VisitIterator` initializer options
    :return: the *{subject: {session: {series: [dicom files]}}}* dictionary
    """
    # Collect the AIRC visits into (subject, session, dicom_files)
    # tuples.
    visits = list(_iter_new_visits(collection, *inputs, **opts))

    # If no images were detected, then bail.
    if not visits:
        logger(__name__).info("No new visits were detected in the input"
                              " directories.")
        return {}
    logger(__name__).debug("%d new visits were detected" % len(visits))

    # Group the DICOM files by series.
    return visits


def _iter_visits(collection, *inputs, **opts):
    """
    Iterates over the visits in the given subject directories.
    Each iteration item is a *(subject, session, series_dict)* tuple,
    formed as follows:

    - The *subject* is the XNAT subject name formatted by
      :data:`SUBJECT_FMT`.

    - The *session* is the XNAT experiment name formatted by
      :data:`SESSION_FMT`.

    - The *series_dict* generator iterates over the files which match the
      :mod:`qipipe.staging.airc_collection` DICOM file include pattern,
      grouped by the scan type.

    The supported AIRC collections are defined by
    :mod:`qipipe.staging.airc_collection`.

    :param collection: the AIRC image collection name
    :param inputs: the subject directories over which to iterate
    :param opts: the :class:`VisitIterator` initializer options
    :yield: iterate over the visit *(subject, session, files)* tuples
    """
    return VisitIterator(collection, *inputs, **opts)


def _iter_new_visits(collection, *inputs, **opts):
    """
    Filters :meth:`qipipe.staging.staging_helper._iter_visits` to iterate
    over the new visits in the given subject directories which are not in XNAT.

    :param collection: the AIRC image collection name
    :param inputs: the subject directories over which to iterate
    :param opts: the :meth:`_iter_visits` options
    :yield: iterate over the new visit *(subject, session, series_dict)* tuples
    """
    opts['filter'] = _is_new_session
    return _iter_visits(collection, *inputs, **opts)


def _is_new_session(subject, session):
    # If the session is not yet in XNAT, then yield the tuple.
    with qixnat.connect() as xnat:
        exists = xnat.get_session(project(), subject, session).exists()
    if exists:
        logger(__name__).debug("Skipping the %s %s %s session since it has"
                               " already been loaded to XNAT." %
                               (project(), subject, session))

    return not exists


def _group_visits(*visit_tuples):
    """
    Creates the staging dictionary for the new images in the given
    visit tuples.

    :param visit_tuples: the *(subject, session, {scan type: [dicom files]})*
        tuples to group
    :return: the *{subject: {session: {scan type: {series: [dicom files]}}}}*
        dictionary
    """
    # The {subject: {session: {series: {scan_type: [dicom files]}}}} output.
    stg_dict = nested_defaultdict(dict, 3)
    for sbj, sess, scan_type_dict in visit_tuples:
        # Group the session DICOM input files by series within scan type.
        for scan_type, dcm_iter in scan_type_dict.iteritems():
            series_dict = group_dicom_files_by_series(dcm_iter)
            for series, dcm_iter in series_dict.iteritems():
                stg_dict[sbj][sess][scan_type][series] = dcm_iter

    return stg_dict


class VisitIterator(object):

    """
    **VisitIterator** is a generator class for AIRC visits.
    """

    def __init__(self, collection, *subject_dirs, **opts):
        """
        :param collection: the AIRC image collection name
        :param subject_dirs: the subject directories over which to iterate
        :param opts: the following initialization options:
        :keyword filter: a *(subject, session)* selection filter
        """
        self.collection = airc.collection_with_name(collection)
        """The AIRC collection with the given name."""

        self.subject_dirs = subject_dirs
        """The input directories."""

        self.filter = opts.get('filter', lambda subject, session: True)
        """The (subject, session) selection filter."""

    def __iter__(self):
        return self.next()

    def next(self):
        """
        Iterates over the visits in the subject directories.
        
        :yield: the next (subject, session, {scan type: DICOM files}) tuple
        """
        # The visit subdirectory match pattern.
        vpat = self.collection.session_pattern
        
        # The DICOM file search pattern depends on the scan type.
        dcm_pat_dict = self.collection.dcm_pat_dict
        logger(__name__).debug("The DICOM file search pattern is %s..." %
                               dcm_pat_dict)
        
        # Iterate over the visits.
        with qixnat.connect():
            for sbj_dir in self.subject_dirs:
                sbj_dir = os.path.abspath(sbj_dir)
                logger(__name__).debug("Discovering sessions in %s..." %
                                       sbj_dir)
                
                # Make the XNAT subject name.
                sbj_nbr = self.collection.path2subject_number(sbj_dir)
                sbj = SUBJECT_FMT % (self.collection.name, sbj_nbr)
                # The subject subdirectories which match the visit pattern.
                sess_matches = [subdir for subdir in os.listdir(sbj_dir)
                                if re.match(vpat, subdir)]
                
                # Generate the new (subject, session, DICOM files) items in
                # each visit.
                for sess_subdir in sess_matches:
                    # The visit directory path.
                    sess_dir = os.path.join(sbj_dir, sess_subdir)
                    # Silently skip non-directories.
                    if os.path.isdir(sess_dir):
                        # The visit (session) number.
                        sess_nbr = self.collection.path2session_number(
                            sess_subdir)
                        # The XNAT session name.
                        sess = SESSION_FMT % sess_nbr
                        
                        # Apply the selection filter, e.g. an XNAT existence
                        # check. If the session passes the filter, then
                        # the files qualify for iteration.
                        if self.filter(sbj, sess):
                            logger(__name__).debug("Discovered session %s in"
                                                   " %s" % (sess, sess_dir))
                            # The DICOM file match patterns.
                            scan_type_dict = {}
                            for scan_type, dcm_pat in dcm_pat_dict.iteritems():
                                file_pat = os.path.join(sess_dir, dcm_pat)
                                # The visit directory DICOM file iterator.
                                dcm_file_iter = glob.iglob(file_pat)
                                scan_type_dict[scan_type] = dcm_file_iter
                            yield (sbj, sess, scan_type_dict)



import os
from matplotlib import (pyplot, colors, cm)
import numpy as np
import nibabel as nib
from qiutil.logging import logger

def normalize(value, vmin, vspan):
    """
    Maps the given input value to [0, 1].
    
    :param value: the input value
    :param vmin: the minimum input range value
    :param vspan: the value range span (maxium - minimum)
    @return (*in_val* - *vmin*) / *vspan*
    """
    return (value - vmin) / vspan


def discretize(in_file, out_file, nvalues, start=0, threshold=None,
               normalizer=normalize):
    """
    Transforms the given input image file to an integer range with
    the given number of values. The range starts at the given start
    value. The input values are uniformly mapped into the output
    range. For example, if the input values range from 0.0 to 3.0
    *nvalues* is 101, and the start is 0, then an input value of 0
    is transformed to 0, 3.0 is transformed to 100, and the intermediate
    input values are proportionately transformed to the output range.
    
    If a threshold is specified, then every input value which maps to
    an output value less than (*threshold* * *nvalues*) - *start* is
    transformed to the output start value. For example, if the input
    values range from 0.0 to 3.0, then::
    
        discretize(in_file, out_file, 1001, threshold=0.5)
    
    transforms input values as follows:
    
    * If the input value maps to the first half of the output range,
      then the output value is 0.
    
    * Otherwise, the input value *v* maps to the output value
      (*v* * 1000) / 3.
    
    :param in_file: the input file path
    :param out_file: the output file path
    :param nvalues: the number of output entries
    :param start: the starting output value (default 0)
    :param threshold: the threshold in the range start to nvalues
      (default start)
    :param normalize: an optional function to normalize the input
      value (default :meth:`normalize`)
    """
    # If there is an output file argument, then ensure that there is an
    # output parent directory. Otherwise, use the default output file.
    
    # The logger.
    log = logger(__name__)
    
    # The default threshold is 0.
    if not threshold:
        threshold = start
    # Validate the threshold.
    if threshold < start or threshold >= start + nvalues:
        raise ValueError("The threshold is not in the color range [%d-%d]: %d" %
                         (start, start + nvalues, threshold))
    
    print ("Color LUT start: %d end: %d threshold: %d" %
           (start, start + nvalues - 1, threshold))
    
    # Load the NiFTI image.
    in_img = nib.load(in_file)
    print ("Loaded %s." % in_file)
    # The image data 3D array.
    in_data = in_img.get_data()
    
    # Compute the minimum and maximum value.
    vmin = float('infinity')
    vmax = float('-infinity')
    # Each data x value is a [y][z] 2D array.
    for yz in in_data:
        # Iterate over the z arrays.
        for z in yz:
            vmin = min(vmin, *z)
            vmax = max(vmax, *z)

    print ("Computed value minimum %f." % vmin)
    print ("Computed value maximum %f." % vmax)

    # The maximum offset is one less than the number of values.
    max_offset = nvalues - 1
    # The length of the value range.
    vspan = vmax - vmin
    # Track the value count by decile.
    decile_cnts = [0] * 10
    
    # The target image data is an array of shorts with the
    # same shape as the input data.
    shape = in_data.shape
    out_data = np.empty(shape, dtype=np.int16)
    
    # Normalize the values into the discrete range.
    for i in range(shape[0]):
        for j in range(shape[1]):
            for k in range(shape[2]):
                # The input voxel value.
                in_val = in_data[i][j][k]
                # The proportional offset into the range.
                proportion = normalizer(in_val, vmin, vspan)
                # The offset in the [0..nvalues - 1] range.
                offset = int(round(proportion * max_offset))
                # The output value.
                out_val = start + offset
                # Check against the threshold.
                if out_val < threshold:
                    offset = 0
                    out_val = start
                # Set the output voxel value.
                out_data[i][j][k] = out_val
                # Increment the output value decile count.
                di = (offset * 10) / nvalues
                decile_cnts[di] += 1
    
    # Log the total number of values and the decile count.
    value_cnt = reduce(lambda x,y: x * y, shape)
    print ("%d input values were mapped to the inclusive range [%d, %d]." %
           (value_cnt, start, start + nvalues - 1))
    print ("Mapped value decile count: %s." % decile_cnts)
    
    print ("Saving the output as %s..." % out_file)
    hdr = in_img.get_header()
    hdr.set_data_dtype(np.int16)
    out_img = nib.Nifti1Image(out_data, in_img.get_affine(), hdr)
    out_img.to_filename(out_file)



"""
The ``pipeline`` module includes the following QIN workflows:

- :mod:`qipipe.pipeline.qipipeline`: the soup-to-nuts pipeline to stage, mask,
  register and model new images

- :mod:`qipipe.pipeline.staging`: executes the staging workflow to detect new
  images, group them by series, import them into XNAT and prepare them for
  TCIA import

- :mod:`qipipe.pipeline.mask`: creates a mask to subtract extraneous tissue
  from the input images
  
- :mod:`qipipe.pipeline.registration`: masks the target tissue and corrects
  motion artifacts

- :mod:`qipipe.pipeline.modeling`: performs pharmokinetic modeling
"""



import os
import glob
import shutil
from nose.tools import assert_equal

from qipipe.staging.fix_dicom import fix_dicom_headers
from qidicom import reader
from ... import ROOT

# The test fixture.
FIXTURE = os.path.join(ROOT, 'fixtures', 'staging', 'sarcoma', 'Subj_1')

# The test results.
RESULTS = os.path.join(ROOT, 'results', 'staging', 'fix_dicom')

# The collection name.
COLLECTION = 'Sarcoma'

# The new subject.
SUBJECT = 'Sarcoma003'

# The scan type.
SCAN_TYPE = 't1'

class TestFixDicom(object):

    """Fix DICOM header unit tests."""

    def test_fix_dicom_headers(self):
        shutil.rmtree(RESULTS, True)
        dest = os.path.dirname(RESULTS)
        fixed = fix_dicom_headers(COLLECTION, SUBJECT, FIXTURE, dest=dest)
        # Verify the result.
        for ds in reader.iter_dicom(*fixed):
            assert_equal(ds.BodyPartExamined, 'CHEST',
                         "Incorrect Body Part: %s" % ds.BodyPartExamined)
            assert_equal(
                ds.PatientID, SUBJECT, "Incorrect Patient ID: %s" % ds.PatientID)
        # Cleanup.
        shutil.rmtree(RESULTS, True)


if __name__ == "__main__":
    import nose
    nose.main(defaultTest=__name__)
